import os
import sys
from typing import Optional, Union

def get_win_install_parent_path() -> str:
    return os.path.join('C:', os.sep, 'Program Files (x86)')

def get_win_install_path(cst_version: int) -> str:
    return os.path.join(get_win_install_parent_path(), f'CST Studio Suite {cst_version}')

def find_win_installed_versions() -> list[str]:
    res = []
    dir_path = get_win_install_parent_path()
    for file_path in os.listdir(dir_path):
        if os.path.isdir(os.path.join(dir_path, file_path)):
            if file_path.startswith('CST Studio Suite '):
                res.append(file_path)
    return tuple(map(lambda x: int(x[17:]), res))

def get_win_latest_installed_version() -> Optional[int]:
    versions = find_win_installed_versions()
    return None if len(versions) == 0 else max(versions)

def add_win_installation_to_path_by_version(version: int):
    add_win_installation_to_path(get_win_install_path(version))

def add_win_installation_to_path(cst_install_dir: int):
    path = os.path.join(cst_install_dir, 'AMD64', 'python_cst_libraries')
    if os.path.isdir(path):
        sys.path.append(path)
    else:
        raise RuntimeError(
            'Failed to add the CST Python lib folder to path. '+
            f'Directory \'{path}\' doesn\'t exist.')

def load_win_cst_python_lib(version_or_install_dir: Optional[Union[int,str]] = None) -> None:
    if version_or_install_dir is None:
        version_or_install_dir = get_win_latest_installed_version()
        if version_or_install_dir is None:
            raise RuntimeError(
                'No CST Studio Suite installation directory found at standard paths. '+
                'Please provide the installation directory manually.')
    if isinstance(version_or_install_dir, int):
        add_win_installation_to_path_by_version(version_or_install_dir)
    else:
        add_win_installation_to_path(version_or_install_dir)

if __name__ == '__main__':
    print(get_win_install_parent_path())
    print(find_win_installed_versions())


from . import IVBAProvider, VBAObjWrapper

class STL(VBAObjWrapper):
    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'STL')

    def reset(self):
        self.invoke_method('Reset')

    def set_file_name(self, file_name: str):
        self.invoke_method('FileName', file_name)

    def set_name(self, name: str):
        self.invoke_method('Name', name)

    def set_component(self, name: str):
        self.invoke_method('Component', name)

    def set_scale_to_unit(self, flag: bool = True):
        self.invoke_method('ScaleToUnit', flag)

    def set_import_file_units(self, units: str):
        self.invoke_method('ImportFileUnits', units)

    def set_export_from_active_coordinate_system(self, flag: bool = True):
        self.invoke_method('ExportFromActiveCoordinateSystem', flag)

    def set_import_to_active_coordinate_system(self, flag: bool = True):
        self.invoke_method('ImportToActiveCoordinateSystem', flag)

    def set_export_file_units(self, units: str):
        self.invoke_method('ExportFileUnits', units)

    def set_normal_tolerance(self, tolerance: float):
        self.invoke_method('NormalTolerance', tolerance)

    def set_surface_tolerance(self, tolerance: float):
        self.invoke_method('SurfaceTolerance', tolerance)

    def read(self):
        self.invoke_method('Read')

    def write(self):
        self.invoke_method('Write')


import win32com.client
import os.path
import shutil

class Interface:
    def clear_cache():
        cache_path = os.path.join(os.getenv('TMP'), 'gen_py')
        shutil.rmtree(cache_path, True)

    def __init__(self, version=None):
        self.stored_quiet_mode = False
        if(version is None):
            self.__open_app_latest_version()
        else:
            self.__open_app_specific_version(version)
        self.enable_quiet_mode()

    def __del__(self):
        self.disable_quiet_mode()

    def __open_app_latest_version(self):
        self.com_object = win32com.client.gencache\
            .EnsureDispatch('CSTStudio.Application')

    def __open_app_specific_version(self, version):
        self.com_object = win32com.client.gencache\
            .EnsureDispatch('CSTStudio.Application.'+str(version))

    def store_quiet_mode(self):
        self.stored_quiet_mode = self.is_in_quiet_mode()

    def restore_quiet_mode(self):
        self.set_quiet_mode(self.stored_quiet_mode)

    def store_and_disable_quiet_mode(self):
        self.store_quiet_mode()
        self.disable_quiet_mode()

    def store_and_enable_quiet_mode(self):
        self.store_quiet_mode()
        self.enable_quiet_mode()

    def open_3d_project(self, path):
        self.com_object.OpenFile(os.path.abspath(path))

    def open_ds_project(self, path):
        self.com_object.OpenFile(os.path.abspath(path))

    def create_new_mws_project(self):
        # Creates a new CST Microwave Studio project.
        self.com_object.NewMWS()

    def create_new_ems_project(self):
        # Creates a new CST EM Studio project.
        self.com_object.NewEMS()

    def create_new_ps_project(self):
        # Creates a new CST Particle Studio project.
        self.com_object.NewPS()

    def create_new_mps_project(self):
        # Creates a new CST Mphysics Studio project.
        self.com_object.NewMPS()

    def create_new_pcbs_project(self):
        # Creates a new CST PCB Studio project.
        self.com_object.NewPCBS()

    def create_new_ds_project(self):
        # Creates a new CST Design Studio project.
        self.com_object.NewDS()

    def enable_quiet_mode(self, flag: bool = True):
        self.com_object.SetQuietMode(flag)
        self.in_quiet_mode = flag

    def disable_quiet_mode(self):
        self.com_object.SetQuietMode(False)
        self.in_quiet_mode = False

    def set_quiet_mode(self, enabled=True):
        self.com_object.SetQuietMode(enabled)
        self.in_quiet_mode = enabled

    def is_in_quiet_mode(self):
        return self.in_quiet_mode

    def close(self):
        self.com_object.Quit()


from . import IVBAProvider, VBAObjWrapper

class MeshShapes(VBAObjWrapper):
    def __init__(self, vbap: IVBAProvider) -> None:
        super.__init__(vbap, 'MeshShapes')

    def reset(self):
        self.record_method('Reset')

    def delete(self, mesh_shape_name: str):
        self.record_method('Delete', mesh_shape_name)

    def rename(self, old_mesh_shape_name: str, new_mesh_shape_name: str):
        self.record_method('Rename', old_mesh_shape_name, new_mesh_shape_name)

    def create_folder(self, name: str):
        self.record_method('NewFolder', name)

    def delete_folder(self, name: str):
        self.record_method('DeleteFolder', name)

    def rename_folder(self, old_name: str, new_name: str):
        self.record_method('RenameFolder', old_name, new_name)

    def change_material(self, mesh_shape_name: str, material_name: str):
        self.record_method('ChangeMaterial', mesh_shape_name, material_name)

    def add_name(self, element_name: str):
        self.record_method('AddName', element_name)

    def delete_multiple(self):
        self.record_method('DeleteMultiple')

    def set_tolerance(self, value: float):
        self.record_method('Tolerance', value)

    def resolve_intersections(self):
        self.record_method('ResolveIntersections')

    def set_mesh_element_size(self, value: float):
        self.record_method('MeshElementSize', value)

    def create_mesh_shapes_by_facetting(self):
        self.record_method('CreateMeshShapesByFacetting')

    def create_mesh_shapes_by_remeshing(self):
        self.record_method('CreateMeshShapesByRemeshing')


from . import Project
from . import ComObjectWrapper

class Align(ComObjectWrapper):
    ALIGN_WHAT_SHAPE = 'Shape'
    ALIGN_WHAT_SUBPROJECT = 'Subproject'
    ALIGN_WHAT_PART = 'Part'
    ALIGN_WHAT_MIXED = 'Mixed'

    ALIGN_HOW_FACES = 'Faces'
    ALIGN_HOW_ROTATE = 'Rotate'
    ALIGN_HOW_ROTATE_BY_ANGLE = 'RotateByDegree'

    PICK_WHAT_SOURCE_PLANE = 'SourcePlane'
    PICK_WHAT_TARGET_PLANE = 'TargetPlane'
    PICK_WHAT_ZERO_ANGLE = 'ZeroAngle'
    PICK_WHAT_FINAL_ANGLE = 'FinalAngle'

    PICK_KIND_FACE = 'Face'
    PICK_KIND_EDGE = 'Edge'
    PICK_KIND_POINT = 'Point'

    NUMERICAL_VALUE_ANGLE = 'Angle'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Align

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_name(self, name: str):
        self.invoke_method('Name', name)

    def add_name(self, name: str):
        self.invoke_method('AddName', name)

    def align(self, what: str, how: str):
        self.invoke_method('Align', what, how)

    def set_kind_of_pick_for(self, what: str, kind: str):
        self.invoke_method('SetKindOfPickFor', what, kind)

    def set_name_to_step(self, what: str, name: str):
        self.invoke_method('SetNameToStep', what. name)

    def set_numerical_value(self, what: str, value: float):
        self.invoke_method('SetNumericalValue', what, value)

    def set_opposite_face_orientation(self, flag: bool = True):
        self.invoke_method('SetOppositeFaceOrientation', flag)

    def set_clear_subproject_import_info(self, flag: bool = True):
        self.invoke_method('ClearSubProjectImportInfo', flag)

    #TODO: implement methods {specializations of enum values}


import numpy as np

DEFAULT_Z_REF = 50

def convert_s_to_s(s_mat, z_ref: complex, z_ref_new: complex):
    E = np.eye(s_mat.shape[0])
    r = (z_ref_new - z_ref)/(z_ref_new + z_ref)
    r_mat = E*r
    a = np.sqrt(1-np.abs(r)**2)*np.sign(1-r)
    a_mat = E*a
    a_mat_inv = E/a
    return a_mat_inv@(s_mat - r_mat)@np.linalg.inv((E - r_mat@s_mat))@a_mat

def convert_s_to_z(s_mat, z_ref: complex = DEFAULT_Z_REF):
    E = np.eye(s_mat.shape[0])
    sqrt_z_ref_mat = E*np.sqrt(z_ref)
    return sqrt_z_ref_mat@np.linalg.solve(E-s_mat, E+s_mat)@sqrt_z_ref_mat

def convert_z_to_s(z_mat, z_ref: complex = DEFAULT_Z_REF):
    E = np.eye(z_mat.shape[0])
    z_ref_mat = z_ref*E
    return np.linalg.solve(z_mat + z_ref_mat, z_mat - z_ref_mat)

def convert_s_to_y(s_mat, z_ref: complex = DEFAULT_Z_REF):
    E = np.eye(s_mat.shape[0])
    return (E - s_mat) @ np.linalg.inv(z_ref * (E + s_mat))

def convert_y_to_s(y_mat, z_ref: complex = DEFAULT_Z_REF):
    E = np.eye(y_mat.shape[0])
    return np.linalg.solve(E + z_ref*y_mat, E - z_ref*y_mat)

def convert_s_to_abcd(s_mat, z_ref: complex = DEFAULT_Z_REF):
    s11, s12, s21, s22 = s_mat[0][0], s_mat[0][1], s_mat[1][0], s_mat[1][1]
    k = 2*s21
    l = s12*s21
    a = ((1 + s11)*(1 - s22) + l)/k
    b = ((1 + s11)*(1 + s22) - l)/k*z_ref
    c = ((1 - s11)*(1 - s22) - l)/k/z_ref
    d = ((1 - s11)*(1 + s22) + l)/k
    return np.array([[a, b], [c, d]])

def convert_abcd_to_s(abcd_mat, z_ref: complex = DEFAULT_Z_REF):
    a, b, c, d = abcd_mat[0][0], abcd_mat[0][1], abcd_mat[1][0], abcd_mat[1][1]
    k = a + b/z_ref + c*z_ref + d
    s11 = (a + b/z_ref - c*z_ref - d)/k
    s12 = 2*(a*d - b*c)/k
    s21 = 2/k
    s22 = (-a + b/z_ref - c*z_ref + d)/k
    return np.array([[s11, s12], [s21, s22]])

def convert_s_to_h(s_mat, z_ref: complex = DEFAULT_Z_REF):
    s11, s12, s21, s22 = s_mat[0][0], s_mat[0][1], s_mat[1][0], s_mat[1][1]
    delta = (1 - s11)*(1 + s22) + s12*s21
    h11 = ((1 + s11)*(1 + s22) - s12*s21)*z_ref/delta
    h12 = 2*s12/delta
    h21 = -2*s21/delta
    h22 = (((1 - s11)*(1 - s22) - s12*s21))/z_ref/delta
    return np.array([[h11, h12], [h21, h22]])

def convert_h_to_s(h_mat, z_ref: complex = DEFAULT_Z_REF):
    h11, h12, h21, h22 = h_mat[0][0]/z_ref, h_mat[0][1], h_mat[1][0], h_mat[1][1]*z_ref
    delta = (h11 + 1)*(h22 + 1) - h12*h21
    s11 = ((h11 - 1)*(h22 + 1) - h12*h21)/delta
    s12 = 2*h12/delta
    s21 = -2*h21/delta
    s22 = ((1 + h11)*(1 - h22) + h12*h21)/delta
    return np.array([[s11, s12], [s21, s22]])


from . import IVBAProvider, VBAObjWrapper

class Shape(VBAObjWrapper):
    def __init__(self, vbap: IVBAProvider, shape_name: str) -> None:
        super.__init__(vbap, shape_name)

    def reset(self):
        self.record_method('Reset')

    def set_name(self, name: str):
        self.record_method('Name', name)

    def set_component(self, component_name: str):
        self.record_method('Component', component_name)

    def set_material(self, mat_name: str):
        self.record_method('Material', mat_name)

    def create(self):
        self.record_method('Create')


from enum import Enum

class VBATypeName(Enum):
    Byte = 'Byte'
    Integer = 'Integer'
    Long = 'Long'
    String = 'String'
    Single = 'Single'
    Double = 'Double'
    Boolean = 'Boolean'
    Variant = 'Variant'


from ..CST import ResultTree
import re

#TODO: modify to use custom farfield names instead

class Farfields:
    ITEM_ROOT = 'Farfields'
    ITEM_REGEX = '^Farfields\\\\farfield \\(f=(\d+(?:\.\d*)?|\.\d+)\\) \\[(\d+)\\]$'
    ITEM_TEMPL = 'Farfields\\farfield (f={}) [{}]'

    def get_name(frequency: float, port: int = 1) -> str:
        return Farfields.get_item_name_from_params((frequency, port))

    def get_params_from_item_name(name):
        x = re.search(Farfields.ITEM_REGEX, name)
        return None if x is None else (float(x.group(1)), int(x.group(2)))

    def get_item_name_from_params(params):
        frequency = float(params[0]) if params[0] != int(params[0]) else int(params[0])
        port = int(params[1])
        return Farfields.ITEM_TEMPL.format(frequency, port)

    def __init__(self, result_tree: ResultTree) -> None:
        self.result_tree = result_tree
        self.project = result_tree.project
        self.com_object = result_tree.com_object

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def are_available(self):
        return self.result_tree.has_item(Farfields.ITEM_ROOT) and self.__has_at_least_one_result()

    def get_available_params(self):
        res = map(Farfields.get_params_from_item_name, self.__get_available_items())
        res = Farfields.__remove_none_values(res)
        return res

    def get_by_params(self, params):
        if params in self.get_available_params():
            return Farfield(self, params[0], params[1])
        return None

    def get_all(self):
        map_fn = lambda params: Farfield(self, params[0], params[1])
        return list(map(map_fn, self.get_available_params()))

    def __has_at_least_one_result(self):
        return len(self.result_tree.get_first_child_name(Farfields.ITEM_ROOT)) > 0

    def __get_available_items(self):
        res = []
        child_name = self.result_tree.get_first_child_name(Farfields.ITEM_ROOT)
        while len(child_name) > 0:
            res.append(child_name)
            child_name = self.result_tree.get_next_item_name(child_name)
        return res

    def __remove_none_values(arr):
        return list(filter(lambda x: x is not None, arr))

class Farfield:
    def __init__(self, farfields: Farfields, frequency: float, port: int) -> None:
        self.farfields = farfields
        self.project = farfields.project
        self.frequency = frequency
        self.port = port

    def select(self):
        item_name = Farfields.get_item_name_from_params((self.frequency, self.port))
        self.project.select_tree_item(item_name)

    def get_frequency(self) -> float:
        return self.frequency

    def get_port(self) -> int:
        return self.port


from . import Project
from . import ComObjectWrapper

class ADSCosimulation(ComObjectWrapper):
    SOLVER_TYPE_TRANSIENT = 'transient'
    SOLVER_TYPE_FREQUNCY_DOMAIN = 'frequency domain'

    PARAM_TYPE_NONE = 'None'
    PARAM_TYPE_LENGTH = 'Length'
    PARAM_TYPE_TEMPERATURE = 'Temperature'
    PARAM_TYPE_VOLTAGE = 'Voltage'
    PARAM_TYPE_CURRENT = 'Current'
    PARAM_TYPE_RESISTANCE = 'Resistance'
    PARAM_TYPE_CONDUCTANCE = 'Conductance'
    PARAM_TYPE_CAPACITANCE = 'Capacitance'
    PARAM_TYPE_INDUCTANCE = 'Inductance'
    PARAM_TYPE_FREQUENCY = 'Frequency'
    PARAM_TYPE_TIME = 'Time'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.ADSCosimulation

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def set_cosimulation_enabled(self, flag: bool = True):
        self.invoke_method('EnableCoSimulation', flag)

    def set_use_interpolation(self, flag: bool = True):
        self.invoke_method('UseInterpolation', flag)

    def set_solver_type(self, solver_type: str):
        self.invoke_method('SolverType', solver_type)

    def set_description(self, description: str):
        self.invoke_method('Description', description)

    def set_parameter_info(
            self, param_name: str, use: bool, param_type: str, nominal_value: float,
            step_size: float):
        self.invoke_method(
            'ParameterInformation', param_name, use, param_type, nominal_value, step_size)


from ..CST import FarfieldPlot, IVBAProvider
from .. import FFS
import numpy as np
import pathlib
import os.path

class ASCIIFarfieldExporter:
    DEFAULT_STEP_DEG: float = 5.0

    def __init__(self) -> None:
        self.coord_system = FarfieldPlot.COORD_SYSTEM_SPHERICAL
        self.plot_mode = FarfieldPlot.PLOT_MODE_EFIELD
        self.polar_angle_step = ASCIIFarfieldExporter.DEFAULT_STEP_DEG
        self.lateral_angle_step = ASCIIFarfieldExporter.DEFAULT_STEP_DEG
        self.ascii_version = FarfieldPlot.ASCII_VERSION_2010
        self.origin = (0.0, 0.0, 0.0)
        self.radius = 1.0
        self.current_ffs: FFS.Farfield = None

    def set_coord_system(self, coord_system: str):
        self.coord_system = coord_system

    def set_plot_mode(self, plot_mode: str):
        self.plot_mode = plot_mode

    def set_polar_angle_step_deg(self, step_deg: float):
        self.polar_angle_step = step_deg

    def set_polar_angle_step_rad(self, step_rad: float):
        self.set_polar_angle_step_deg(np.rad2deg(step_rad))

    def set_lateral_angle_step_deg(self, step_deg: float):
        self.lateral_angle_step = step_deg

    def set_lateral_angle_step_rad(self, step_rad: float):
        self.set_lateral_angle_step_deg(np.rad2deg(step_rad))

    def set_ascii_version(self, ascii_version: str):
        self.ascii_version = ascii_version

    def set_origin(self, x: float, y: float, z: float):
        self.origin = (x, y, z)

    def set_radius(self, radius: float):
        self.radius = radius

    def prepare(self, vbap: IVBAProvider, farfield_name: str):
        self.vbap = vbap
        self.farfield_name = farfield_name
        self.__select_tree_item(farfield_name)
        # self.__prepare_ffplot() #TODO: fix this
        self.__export_file()

    def export_abs(self):
        return None

    def export_complex_theta(self):
        return self.current_ffs.get_theta_component()

    def export_complex_phi(self):
        return self.current_ffs.get_phi_component()

    def __select_tree_item(self, farfield_name: str):
        self.vbap.get_quiet_mode_controller().store_and_disable_quiet_mode()
        self.vbap.invoke_function('SelectTreeItem', f'Farfields\\{farfield_name}')
        self.vbap.get_quiet_mode_controller().restore_quiet_mode()

    def __get_tmp_path(self) -> str:
        return self.vbap.query_function('GetProjectPath', 'Temp')

    def __prepare_ffplot(self):
        ffplot = FarfieldPlot(self.vbap)
        ffplot.reset()
        ffplot.set_plot_mode(self.plot_mode)
        ffplot.set_plot_type(FarfieldPlot.PLOT_TYPE_3D)
        ffplot.set_coord_system(self.coord_system)
        ffplot.set_db_unit(FarfieldPlot.UNIT_CODE_0) # basic units
        ffplot.set_origin(self.origin[0], self.origin[1], self.origin[2])
        ffplot.set_polarization_linear()
        ffplot.set_lock_steps(False)
        ffplot.set_theta_step_deg(self.polar_angle_step)
        ffplot.set_phi_step_deg(self.lateral_angle_step)
        ffplot.set_virtual_sphere_radius(self.radius)
        ffplot.set_scale_linear()
        ffplot.set_ascii_export_version(self.ascii_version)
        ffplot.plot()

    def __export_file(self):
        ffplot = FarfieldPlot(self.vbap)
        tmp_path = self.__get_tmp_path()
        file_path = os.path.join(tmp_path, 'tmp_ff.ffs')
        ffplot.export_source_as_ascii(file_path)
        self.__load_ffs(file_path)
        os.remove(file_path)

    def __load_ffs(self, path):
        parser = FFS.Parser()
        parser.parse_string(pathlib.Path(path).read_text())
        self.current_ffs = parser.get_farfields()[0]


from . import Project
from . import ComObjectWrapper

class PostProcess1D(ComObjectWrapper):
    APPLY_TARGET_S_PARAM = 'S-parameter'
    APPLY_TARGET_PROBES = 'Probes'
    APPLY_TARGET_MONITORS = 'Monitors'

    OPERATION_TYPE_TIME_WINDOW = 'Time Window'
    OPERATION_TYPE_AR_FILTER = 'AR-Filter'
    OPERATION_TYPE_PHASE_DEEMBEDDING = 'Phase Deembedding'
    OPERATION_TYPE_RENORMALIZATION = 'Renormalization'
    OPERATION_TYPE_VSWR = 'VSWR'
    OPERATION_TYPE_YZ_MATRICES = 'YZ-matrices'
    OPERATION_TYPE_EXCLUDE_PORT_MODES = 'Exclude Port Modes'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.PostProcess1D

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_apply_target(self, apply_target: str):
        self.invoke_method('ApplyTo', apply_target)

    def add_operation(self, operation_type: str):
        self.invoke_method('AddOperation', operation_type)

    def set_deembed_distance(self, port_name: int, distance: float):
        self.invoke_method('SetDeembedDistance', port_name, distance)

    def set_renorm_impedance(self, port_name: int, mode_name: int, impedance: float):
        self.invoke_method('SetRenormImpedance', port_name, mode_name, impedance)

    def set_renorm_impedance_on_all_ports(self, impedance: float):
        self.invoke_method('SetRenormImpedanceOnAllPorts', impedance)

    def reset_renorm_impedance_on_all_ports(self):
        self.invoke_method('SetUnnormImpedanceOnAllPorts')

    def set_consider_port_mode(self, port_name: int, mode_name: int, flag: bool = True):
        self.invoke_method('SetConsiderPortMode', port_name, mode_name, flag)

    def run(self):
        self.invoke_method('Run')

    def set_operation_active(self, operation_type: str, active: bool = True):
        self.invoke_method('ActivateOperation', operation_type, active)


from . import Project
from . import ComObjectWrapper

class Arc(ComObjectWrapper):
    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Arc

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_name(self, arc_name: str):
        self.invoke_method('Name', arc_name)

    def set_curve(self, curve_name: str):
        self.invoke_method('Curve', curve_name)

    def set_orientation_clockwise(self):
        self.invoke_method('Orientation', 'Clockwise')

    def set_orientation_counter_clockwise(self):
        self.invoke_method('Orientation', 'CounterClockwise')

    def set_center_x(self, x: float):
        self.invoke_method('Xcenter', x)

    def set_center_y(self, y: float):
        self.invoke_method('Ycenter', y)

    def set_center(self, x: float, y: float):
        self.set_center_x(x)
        self.set_center_y(y)

    def set_start_point_x(self, x: float):
        self.invoke_method('X1', x)

    def set_start_point_y(self, y: float):
        self.invoke_method('Y1', y)

    def set_start_point_x(self, x: float, y: float):
        self.set_start_point_x(x)
        self.set_start_point_y(y)


from .CST import Parameters
from typing import Callable, Union

numeric_value = Union[float, int]

class ParameterSweep:
    def __init__(self, project_params: Parameters) -> None:
        self.project_params = project_params
        self.parameters = {}
        self.cb_before_it = None
        self.cb_after_it = None
        self.cb_start_sim = None
        self.curr_param_list = []
        self.param_indices = {}
        self.all_combinations_simulated = False

    def set_parameter(self, param_name: str, values: list[numeric_value]):
        self.parameters[param_name] = values

    def delete_parameter(self, param_name: str):
        self.parameters.pop(param_name)

    def set_callback_before_iteration(self, callback: Callable):
        self.cb_before_it = callback

    def set_callback_after_iteration(self, callback: Callable):
        self.cb_after_it = callback

    def set_callback_start_simulation(self, callback: Callable):
        self.cb_start_sim = callback

    def start(self):
        self.__prepare_sweep()
        while not self.all_combinations_simulated:
            self.__do_next_iteration()

    def __do_next_iteration(self):
        values = self.__get_current_param_values()
        self.__write_new_param_values(values)
        self.__call_before_it_cb_if_exists(values)
        self.__start_simulation()
        self.__call_after_it_cb_if_exists(values)
        self.__prepare_next_param_combination()

    def __prepare_sweep(self):
        self.__ensure_some_parameters_provided()
        self.all_combinations_simulated = False
        self.curr_param_list = list(self.parameters.keys())
        for param_name in self.curr_param_list:
            self.param_indices[param_name] = 0
        self.__ensure_cb_start_sim_provided()

    def __prepare_next_param_combination(self):
        self.__increment_param_index_if_exists(0)

    def __increment_param_index_if_exists(self, param_index: int):
            if param_index >= len(self.curr_param_list):
                self.all_combinations_simulated = True
            else:
                self.__increment_param_index(param_index)

    def __increment_param_index(self, param_index: int):
            param_name = self.curr_param_list[param_index]
            num_values = len(self.parameters[param_name])
            self.param_indices[param_name] += 1
            if self.param_indices[param_name] == num_values:
                self.param_indices[param_name] = 0
                self.__increment_param_index_if_exists(param_index + 1)

    def __get_current_param_values(self) -> dict[str, numeric_value]:
        values = {}
        for param_name in self.parameters:
            val_index = self.param_indices[param_name]
            param_values = self.parameters[param_name]
            values[param_name] = param_values[val_index]
        return values

    def __write_new_param_values(self, values: dict[str, numeric_value]):
        for param_name in values:
            self.project_params.store(param_name, values[param_name])

    def __start_simulation(self):
        self.cb_start_sim()

    def __call_before_it_cb_if_exists(self, param_values: dict[str, numeric_value]):
        if callable(self.cb_before_it):
            self.cb_before_it(param_values)

    def __call_after_it_cb_if_exists(self, param_values: dict[str, numeric_value]):
        if callable(self.cb_after_it):
            self.cb_after_it(param_values)

    def __ensure_cb_start_sim_provided(self):
        if not callable(self.cb_start_sim):
            raise RuntimeError('No run simulation callback provided.')

    def __ensure_some_parameters_provided(self):
        if not bool(self.parameters):
            raise RuntimeError('No parameters provided to sweep.')


from . import IVBAProvider, VBAObjWrapper, VBATypeName
from typing import Optional

class LumpedElement(VBAObjWrapper):
    TYPE_RLC_PARALLEL = 'rlcparallel'
    TYPE_RLC_SERIAL = 'rlcserial'
    TYPE_DIODE = 'diode'
    TYPE_SPICE_CIRCUIT = 'spicecircuit'
    TYPE_TOUCHSTONE = 'touchstone'
    TYPE_MULTIPIN_GROUP_ITEM = 'multipingroupitem'
    TYPE_MULTIPIN_GROUP_SPICE = 'multipingroupspice'
    TYPE_MULTIPIN_GROUP_TOUCHSTONE = 'multipingrouptouchstone'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'LumpedElement')

    def reset(self):
        self.record_method('Reset')

    def create(self):
        self.record_method('Create')

    def create_multipin(self):
        self.record_method('CreateMultipin')

    def modify(self):
        self.record_method('Modify')

    def set_type(self, elem_type: str):
        self.record_method('SetType', elem_type)

    def set_name(self, name: str):
        self.record_method('SetName', name)

    def set_folder_name(self, folder_name: str):
        self.record_method('Folder', folder_name)

    def set_resistance(self, value: float):
        self.record_method('SetR', value)

    def set_inductance(self, value: float):
        self.record_method('SetL', value)

    def set_capacitance(self, value: float):
        self.record_method('SetC', value)

    def set_blocking_conductivity(self, value: float):
        self.record_method('SetGs', value)

    def set_reverse_current(self, value: float):
        self.record_method('SetI0', value)

    def set_temperature_kelvin(self, value: float):
        self.record_method('SetT', value)

    def set_temperature_celsius(self, value: float):
        self.record_method('SetT', value-273.15)

    def set_temperature_fahrenheit(self, value: float):
        self.record_method('SetT', (value + 459.67)*(5/9))

    def set_point1(self, x: float, y: float, z: float):
        self.record_method('SetP1', False, x, y, z)

    def set_pick_as_point1(self):
        self.record_method('SetP1', True, 0, 0, 0)

    def set_point2(self, x: float, y: float, z: float):
        self.record_method('SetP2', False, x, y, z)

    def set_pick_as_point2(self):
        self.record_method('SetP2', True, 0, 0, 0)

    # returns: (x0, y0, z0, x1, y1, z1) or None
    def get_coordinates(self, name: str) \
            -> Optional[tuple[float, float, float, float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetCoordinates', B, name, D, D, D, D, D, D)
        return None if not res[0] else res[1:]

    # returns: (type, R, L, C, Gs, I0, T, Radius) or None
    def get_properties(self, name: str) \
            -> Optional[tuple[str, float, float, float, float, float, float, float]]:
        B, S, D = VBATypeName.Boolean, VBATypeName.String, VBATypeName.Double
        res = self.query_method_t('GetProperties', B, name, S, D, D, D, D, D, D, D)
        return None if not res[0] else res[1:]

    def set_circuit_file_name(self, file_name: str):
        self.record_method('CircuitFileName', file_name)

    def set_use_relative_path(self, flag: bool = True):
        self.record_method('UseRelativePath', flag)

    def set_circuit_id(self, id: int):
        self.record_method('CircuitId', id)

    def set_use_copy_only(self, flag: bool = True):
        self.record_method('UseCopyOnly', flag)

    def set_invert(self, flag: bool = True):
        self.record_method('SetInvert', flag)

    def set_monitor(self, flag: bool = True):
        self.record_method('SetMonitor', flag)

    def set_radius(self, radius: float):
        self.record_method('SetRadius', radius)

    def set_wire(self, wire: str):
        self.record_method('Wire', wire)

    def set_wire_end_to_end1(self):
        self.record_method('Position', 'end1')

    def set_wire_end_to_end2(self):
        self.record_method('Position', 'end2')

    def set_port_name(self, name: str):
        self.record_method('PortName', name)

    def start_name_iteration(self):
        self.invoke_method('StartLumpedElementNameIteration')

    def connect_multipin_element_pin_to_sub_element(
            self, multipin_name: str, circuit_pin_name: str, multipin_sub_element_name: str):
        self.record_method(
            'ConnectMultipinElementPinToSubElement', multipin_name, circuit_pin_name,
            multipin_sub_element_name)

    def connect_multipin_element_pin_to_short(self, multipin_name: str, circuit_pin_name: str):
        self.record_method('ConnectMultipinElementPinToShort', multipin_name, circuit_pin_name)

    def connect_multipin_element_pin_to_open(self, multipin_name: str, circuit_pin_name: str):
        self.record_method('ConnectMultipinElementPinToOpen', multipin_name, circuit_pin_name)

    def get_next_lumped_element_name(self) -> Optional[str]:
        B, S = VBATypeName.Boolean, VBATypeName.String
        res = self.query_method_t('GetNextLumpedElementName', B, S)
        return None if not res[0] else res[1]

    def rename(self, old_name: str, new_name: str):
        self.record_method('Rename', old_name, new_name)

    def delete(self, name: str):
        self.record_method('Delete', name)

    def create_folder(self, name: str):
        self.record_method('NewFolder', name)

    def delete_folder(self, name: str):
        self.record_method('DeleteFolder', name)

    def rename_folder(self, old_name: str, new_name: str):
        self.record_method('RenameFolder', old_name, new_name)


from . import IVBAProvider, VBAObjWrapper

class ProjectUnits(VBAObjWrapper):
    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Units')

    def set_geometry_unit(self, unit: str):
        self.record_method('Geometry', unit)

    def get_geometry_unit(self) -> str:
        return self.query_method_str('GetGeometryUnit')

    def get_geometry_unit_to_si_factor(self) -> float:
        return self.query_method_float('GetGeometryUnitToSI')

    def get_geometry_si_to_unit_factor(self) -> float:
        return self.query_method_float('GetGeometrySIToUnit')

    def set_time_unit(self, unit: str):
        self.record_method('Time', unit)

    def get_time_unit(self) -> str:
        return self.query_method_str('GetTimeUnit')

    def get_time_unit_to_si_factor(self) -> float:
        return self.query_method_float('GetTimeUnitToSI')

    def get_time_si_to_unit_factor(self) -> float:
        return self.query_method_float('GetTimeSIToUnit')

    def set_frequency_unit(self, unit: str):
        self.record_method('Frequency', unit)

    def get_frequency_unit(self) -> str:
        return self.query_method_str('GetFrequencyUnit')

    def get_frequency_unit_to_si_factor(self) -> float:
        return self.query_method_float('GetFrequencyUnitToSI')

    def get_frequency_si_to_unit_factor(self) -> float:
        return self.query_method_float('GetFrequencySIToUnit')

    def set_temperature_unit(self, unit: str):
        self.record_method('TemperatureUnit', unit)

    def get_temperature_unit(self) -> str:
        return self.query_method_str('GetTemperatureUnit')


from . import Project
from . import ComObjectWrapper

class ProjectUnits(ComObjectWrapper):
    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Units

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def set_geometry_unit(self, unit: str):
        self.invoke_method('Geometry', unit)

    def get_geometry_unit(self) -> str:
        return self.invoke_method('GetGeometryUnit')

    def get_geometry_unit_to_si_factor(self) -> float:
        return self.invoke_method('GetGeometryUnitToSI')

    def get_geometry_si_to_unit_factor(self) -> float:
        return self.invoke_method('GetGeometrySIToUnit')

    def set_time_unit(self, unit: str):
        self.invoke_method('Time', unit)

    def get_time_unit(self) -> str:
        return self.invoke_method('GetTimeUnit')

    def get_time_unit_to_si_factor(self) -> float:
        return self.invoke_method('GetTimeUnitToSI')

    def get_time_si_to_unit_factor(self) -> float:
        return self.invoke_method('GetTimeSIToUnit')

    def set_frequency_unit(self, unit: str):
        self.invoke_method('Frequency', unit)

    def get_frequency_unit(self) -> str:
        return self.invoke_method('GetFrequencyUnit')

    def get_frequency_unit_to_si_factor(self) -> float:
        return self.invoke_method('GetFrequencyUnitToSI')

    def get_frequency_si_to_unit_factor(self) -> float:
        return self.invoke_method('GetFrequencySIToUnit')

    def set_temperature_unit(self, unit: str):
        self.invoke_method('TemperatureUnit', unit)

    def get_temperature_unit(self) -> str:
        return self.invoke_method('GetTemperatureUnit')



class IQuietModeController:
    def store_quiet_mode(self) -> None:
        pass

    def restore_quiet_mode(self) -> None:
        pass

    def store_and_disable_quiet_mode(self) -> None:
        pass

    def store_and_enable_quiet_mode(self) -> None:
        pass

    def set_quiet_mode(self, enabled: bool = True) -> None:
        pass

    def is_in_quiet_mode(self) -> bool:
        pass


from . import IVBAProvider, VBAObjWrapper, VBATypeName
import colorsys
from typing import Union, Optional
import numpy as np

class Material(VBAObjWrapper):
    TYPE_PEC = 'PEC'
    TYPE_NORMAL = 'Normal'
    TYPE_ANISOTROPIC = 'Anisotropic'
    TYPE_LOSSY_METAL = 'Lossy Metal'
    TYPE_CORRUGATED_WALL = 'Corrugated wall'
    TYPE_OHMIC_SHEET = 'Ohmic sheet'
    TYPE_TENSOR_FORMULA = 'Tensor formula'
    TYPE_NONLINEAR = 'Nonlinear'

    REF_COORD_SYSTEM_GLOBAL = 'Global'
    REF_COORD_SYSTEM_LOCAL = 'Solid'

    COORD_SYSTEM_CARTESIAN = 'Cartesian'
    COORD_SYSTEM_CYLINDRICAL = 'Cylindrical'
    COORD_SYSTEM_SPHERICAL = 'Spherical'

    TAB_SURF_IMP_MODEL_OPAQUE = 'Opaque'
    TAB_SURF_IMP_MODEL_TRANSPARENT = 'Transparent'

    DISP_FIT_SCHEME_TAB_SI_NTH_ORDER = 'Nth Order'

    ENTRY_BLOCK_UU = 'uu'
    ENTRY_BLOCK_VU = 'vu'
    ENTRY_BLOCK_UV = 'uv'
    ENTRY_BLOCK_VV = 'vv'

    TANGENT_DELTA_MODEL_CONST_SIGMA = 'ConstSigma'
    TANGENT_DELTA_MODEL_CONST_TAN_D = 'ConstTanD'

    CONST_TANGENT_DELTA_STRATEGY_AUTOMATIC_ORDER = 'AutomaticOrder'
    CONST_TANGENT_DELTA_STRATEGY_CUSTOM_ORDER = 'UserDefinedOrder'
    CONST_TANGENT_DELTA_STRATEGY_DJORDJEVIC_SARKAR = 'DjordjevicSarkar'

    DISP_MODEL_EPS_NONE = 'None'
    DISP_MODEL_EPS_DEBYE_1ST_ORDER = 'Debye1st'
    DISP_MODEL_EPS_DEBYE_2ND_ORDER = 'Debye2nd'
    DISP_MODEL_EPS_DRUDE = 'Drude'
    DISP_MODEL_EPS_LORENTZ = 'Lorentz'
    DISP_MODEL_EPS_GENERAL_1ST_ORDER = 'General1st'
    DISP_MODEL_EPS_GENERAL_2ND_ORDER = 'General2nd'
    DISP_MODEL_EPS_GENERAL = 'General'
    DISP_MODEL_EPS_NONLINEAR_2ND_ORDER = 'NonLinear2nd'
    DISP_MODEL_EPS_NONLINEAR_3RD_ORDER = 'NonLinear3rd'
    DISP_MODEL_EPS_NONLINEAR_KERR = 'NonLinearKerr'
    DISP_MODEL_EPS_NONLINEAR_RAMAN = 'NonLinearRaman'

    DISP_MODEL_MU_NONE = 'None'
    DISP_MODEL_MU_DEBYE_1ST_ORDER = 'Debye1st'
    DISP_MODEL_MU_DEBYE_2ND_ORDER = 'Debye2nd'
    DISP_MODEL_MU_DRUDE = 'Drude'
    DISP_MODEL_MU_LORENTZ = 'Lorentz'
    DISP_MODEL_MU_GYROTROPIC = 'Gyrotropic'
    DISP_MODEL_MU_GENERALIZED_DEBYE = 'GeneralizedDebye'
    DISP_MODEL_MU_GENERAL_1ST_ORDER = 'General1st'
    DISP_MODEL_MU_GENERAL_2ND_ORDER = 'General2nd'
    DISP_MODEL_MU_NONLINEAR_2ND_ORDER = 'NonLinear2nd'
    DISP_MODEL_MU_NONLINEAR_3RD_ORDER = 'NonLinear3rd'
    DISP_MODEL_MU_NONLINEAR_KERR = 'NonLinearKerr'
    DISP_MODEL_MU_NONLINEAR_RAMAN = 'NonLinearRaman'

    DISPERSIVE_FITTING_FORMAT_REAL_IMAG = 'Real_Imag'
    DISPERSIVE_FITTING_FORMAT_REAL_TAN_D = 'Real_Tand'

    DISPERSIVE_FITTING_SCHEME_CONDUCTIVITY = 'Conductivity'
    DISPERSIVE_FITTING_SCHEME_1ST_ORDER = '1st Order'
    DISPERSIVE_FITTING_SCHEME_2ND_ORDER = '2nd Order'
    DISPERSIVE_FITTING_SCHEME_NTH_ORDER = 'Nth Order'

    DIRECTION_X = 'x'
    DIRECTION_Y = 'y'
    DIRECTION_Z = 'z'

    SPATIALLY_VAR_MAT_PROP_EPS = 'eps'
    SPATIALLY_VAR_MAT_PROP_MU = 'mu'
    SPATIALLY_VAR_MAT_PROP_EL_SIGMA = 'sigma'
    SPATIALLY_VAR_MAT_PROP_MAG_SIGMA = 'sigmam'

    SPATIALLY_VAR_MAT_MODEL_CONSTANT = 'Constant'
    SPATIALLY_VAR_MAT_MODEL_LUNEBURG = 'Luneburg'
    SPATIALLY_VAR_MAT_MODEL_POWER_LAW = 'PowerLaw'
    SPATIALLY_VAR_MAT_MODEL_GRADED_INDEX = 'GradedIndex'

    SPATIALLY_VAR_MAT_PARAM_CONSTANT_CONSTANT = 'value_constant'
    SPATIALLY_VAR_MAT_PARAM_LUNEBURG_CENTER = 'value_center'
    SPATIALLY_VAR_MAT_PARAM_LUNEBURG_SURFACE = 'value_surface'
    SPATIALLY_VAR_MAT_PARAM_POWER_LAW_AXIS = 'value_axis'
    SPATIALLY_VAR_MAT_PARAM_POWER_LAW_CLADDING = 'value_cladding'
    SPATIALLY_VAR_MAT_PARAM_POWER_LAW_PROFILE = 'value_profile'
    SPATIALLY_VAR_MAT_PARAM_GRADED_INDEX_AXIS = 'value_axis'
    SPATIALLY_VAR_MAT_PARAM_GRADED_INDEX_GRADIENT = 'value_gradient'

    SPACE_MAP_MAT_PROP_EPS = 'eps'
    SPACE_MAP_MAT_PROP_MU = 'mu'
    SPACE_MAP_MAT_PROP_EL_SIGMA = 'sigma'
    SPACE_MAP_MAT_PROP_MAG_SIGMA = 'sigmam'
    SPACE_MAP_MAT_PROP_EPS_INF = 'epsinfinity'
    SPACE_MAP_MAT_PROP_MU_INF = 'muinfinity'
    SPACE_MAP_MAT_PROP_DISP_COEFF0_EPS = 'dispcoeff0eps'
    SPACE_MAP_MAT_PROP_DISP_COEFF0_MU = 'dispcoeff0mu'
    SPACE_MAP_MAT_PROP_DISP_COEFF1_EPS = 'dispcoeff0eps'
    SPACE_MAP_MAT_PROP_DISP_COEFF1_MU = 'dispcoeff0mu'
    SPACE_MAP_MAT_PROP_DISP_COEFF2_EPS = 'dispcoeff0eps'
    SPACE_MAP_MAT_PROP_DISP_COEFF2_MU = 'dispcoeff0mu'
    SPACE_MAP_MAT_PROP_DISP_COEFF3_EPS = 'dispcoeff0eps'
    SPACE_MAP_MAT_PROP_DISP_COEFF3_MU = 'dispcoeff0mu'
    SPACE_MAP_MAT_PROP_DISP_COEFF4_EPS = 'dispcoeff0eps'
    SPACE_MAP_MAT_PROP_DISP_COEFF4_MU = 'dispcoeff0mu'

    SPACE_MAP_MAT_MODEL_3D_CONSTANT = '3DConstant'
    SPACE_MAP_MAT_MODEL_3D_DEFAULT = '3DDefault'
    SPACE_MAP_MAT_MODEL_3D_IMPORT_HEX = '3DImportHex'
    SPACE_MAP_MAT_MODEL_3D_IMPORT_TET = '3DImportTet'

    SPACE_MAP_MAT_PARAM_3D_CONSTANT_CONSTANT = 'value_constant'
    SPACE_MAP_MAT_PARAM_3D_IMPORT_HEX_MAP_FILENAME = 'map_filename'
    SPACE_MAP_MAT_PARAM_3D_IMPORT_TET_MAP_FILENAME = 'map_filename'
    SPACE_MAP_MAT_PARAM_3D_IMPORT_TET_SLIM_MESH_FILENAME = 'slimmesh_filename'

    GENERALIZED_DEBYE_NON_LIN_DEP_NONE = 'None'
    GENERALIZED_DEBYE_NON_LIN_DEP_BH_CURVE = 'BHCurve'
    GENERALIZED_DEBYE_NON_LIN_DEP_TAB_MU_DIFF_CURVE = 'TabMuDiffCurve'

    COATING_TYPE_DEF_PERFECT_ABSORBER = 'PERFECT_ABSORBER'
    COATING_TYPE_DEF_SURFACE_IMPEDANCE_TABLE = 'SURFACE_IMPEDANCE_TABLE'
    COATING_TYPE_DEF_REFLECTION_FACTOR_TABLE = 'REFLECTION_FACTOR_TABLE'
    COATING_TYPE_DEF_REFLECTION_TRANSMISSION_FACTOR_TABLE = 'REFLECTION_TRANSMISSION_FACTOR_TABLE'

    PARTICLE_PROP_NONE = 'None'
    PARTICLE_PROP_SECONDARY_EMISSION = 'SecondaryEmission'
    PARTICLE_PROP_SHEET_TRANSPARENCY = 'SheetTransparency'

    SE_MODEL_NONE = 'None'
    SE_MODEL_FURMAN = 'Furman'
    SE_MODEL_IMPORT = 'Import'
    SE_MODEL_VAUGHAN = 'Vaughan'

    ION_SEE_MODEL_NONE = 'None'
    ION_SEE_MODEL_ION_IMPORT = 'Ion Import'

    PARTICLE_TRANSPARENCY_SETTINGS_SCALAR = 'Scalar'
    PARTICLE_TRANSPARENCY_SETTINGS_IMPORT = 'Import'

    THERMAL_TYPE_PTC = 'PTC'
    THERMAL_TYPE_NORMAL = 'Normal'
    THERMAL_TYPE_ANISOTROPIC = 'Anisotropic'

    SPECIFIC_HEAT_UNIT_J_K_KG = 'J/K/kg'
    SPECIFIC_HEAT_UNIT_KJ_K_KG = 'kJ/K/kg'

    MECHANICS_TYPE_UNUSED = 'Unused'
    MECHANICS_TYPE_ISOTROPIC = 'Isotropic'

    FLOW_RES_PRESSURE_LOSS_TYPE_UVW_BLOCKED = 'Blocked'
    FLOW_RES_PRESSURE_LOSS_TYPE_UVW_COEFFICIENT = 'Coefficient'
    FLOW_RES_PRESSURE_LOSS_TYPE_UVW_CURVE = 'Curve'

    FLOW_RES_PRESSURE_LOSS_TYPE_SHEET_BLOCKED = 'Blocked'
    FLOW_RES_PRESSURE_LOSS_TYPE_SHEET_COEFFICIENT = 'Coefficient'
    FLOW_RES_PRESSURE_LOSS_TYPE_SHEET_CURVE = 'Curve'
    FLOW_RES_PRESSURE_LOSS_TYPE_SHEET_PERFORATION = 'Perforation'
    FLOW_RES_PRESSURE_LOSS_TYPE_SHEET_FREE_AREA_RATIO = 'FreeAreaRatio'

    FLOW_RES_SHAPE_TYPE_HEXAGON = 'Hexagon'
    FLOW_RES_SHAPE_TYPE_CIRCLE = 'Circle'
    FLOW_RES_SHAPE_TYPE_SQUARE = 'Square'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Material')

    def reset(self):
        self.record_method('Reset')

    def create(self):
        self.record_method('Create')

    def set_name(self, name: str):
        self.record_method('Name', name)

    def set_folder_name(self, folder_name: str):
        self.record_method('Folder', folder_name)

    def set_type(self, mat_type: str):
        self.record_method('Type', mat_type)

    def set_frequency_unit(self, unit: str):
        self.record_method('MaterialUnit', 'Frequency', unit)

    def set_geometry_unit(self, unit: str):
        self.record_method('MaterialUnit', 'Geometry', unit)

    def set_time_unit(self, unit: str):
        self.record_method('MaterialUnit', 'Time', unit)

    def set_temperature_unit(self, unit: str):
        self.record_method('MaterialUnit', 'Temperature', unit)

    def delete(self, mat_name: str):
        self.record_method('Delete', mat_name)

    def rename(self, old_name: str, new_name: str):
        self.record_method('Rename', old_name, new_name)

    def create_folder(self, name: str):
        self.record_method('NewFolder', name)

    def delete_folder(self, name: str):
        self.record_method('DeleteFolder', name)

    def rename_folder(self, old_name: str, new_name: str):
        self.record_method('RenameFolder', old_name, new_name)

    def set_color_rgb(self, r: float, g: float, b: float):
        # rgb in range [0, 1]
        self.record_method('Colour', r, g, b)

    def set_color_hsv(self, h: float, s: float, v: float):
        # hsv in range [0, 1]
        r, g, b = colorsys.hsv_to_rgb(h, s, v)
        self.set_color_rgb(r, g, b)

    def set_color_hsl(self, h: float, s: float, l: float):
        # hsl in range [0, 1]
        r, g, b = colorsys.hls_to_rgb(h, l, s)
        self.set_color_rgb(r, g, b)

    def set_transparency(self, transparency: float):
        # The set value is not displayed correctly in the editor settings, but it works.
        self.record_method('Transparency', transparency)

    def set_color_alpha(self, alpha: float):
        self.set_transparency(1.0 - alpha)

    def set_color_rgba(self, r: float, g: float, b: float, alpha: float):
        self.set_color_rgb(r, g, b)
        self.set_color_alpha(alpha)

    def set_color_rgbt(self, r: float, g: float, b: float, transparency: float):
        self.set_color_rgb(r, g, b)
        self.set_transparency(transparency)

    def set_color_hsva(self, h: float, s: float, v: float, alpha: float):
        self.set_color_hsv(h, s, v)
        self.set_color_alpha(alpha)

    def set_color_hsvt(self, h: float, s: float, v: float, transparency: float):
        self.set_color_hsv(h, s, v)
        self.set_transparency(transparency)

    def set_color_hsla(self, h: float, s: float, l: float, alpha: float):
        self.set_color_hsl(h, s, l)
        self.set_color_alpha(alpha)

    def set_color_hslt(self, h: float, s: float, l: float, transparency: float):
        self.set_color_hsl(h, s, l)
        self.set_transparency(transparency)

    def set_wireframe(self, flag: bool = True):
        self.record_method('Wireframe', flag)

    def set_reflection(self, flag: bool = True):
        self.record_method('Reflection', flag)

    def set_allow_outlines(self, flag: bool = True):
        self.record_method('Allowoutline', flag)

    def set_transparent_solid_outlines(self, flag: bool = True):
        self.record_method('Transparentoutline', flag)

    def change_appearance(self):
        self.record_method('ChangeColour')

    def set_rel_permitivity(self, epsilon: float):
        self.record_method('Epsilon', epsilon)

    def set_rel_permitivity_x(self, epsilon: float):
        self.record_method('EpsilonX', epsilon)

    def set_rel_permitivity_y(self, epsilon: float):
        self.record_method('EpsilonY', epsilon)

    def set_rel_permitivity_z(self, epsilon: float):
        self.record_method('EpsilonZ', epsilon)

    def set_rel_permitivity_xyz(self, eps_x: float, eps_y: float, eps_z: float):
        self.set_rel_permitivity_x(eps_x)
        self.set_rel_permitivity_y(eps_y)
        self.set_rel_permitivity_z(eps_z)

    def set_rel_permeability(self, mu: float):
        self.record_method('Mu', mu)

    def set_rel_permeability_x(self, mu: float):
        self.record_method('MuX', mu)

    def set_rel_permeability_y(self, mu: float):
        self.record_method('MuY', mu)

    def set_rel_permeability_z(self, mu: float):
        self.record_method('MuZ', mu)

    def set_rel_permeability_xyz(self, mu_x: float, mu_y: float, mu_z: float):
        self.set_rel_permeability_y(mu_y)
        self.set_rel_permeability_x(mu_x)
        self.set_rel_permeability_z(mu_z)

    def set_material_density(self, rho_kg_per_m3: float):
        self.record_method('Rho', rho_kg_per_m3)

    def add_coated_material(self, mat_name: str, thickness: float):
        self.record_method('AddCoatedMaterial', mat_name, thickness)

    def set_reference_coord_system(self, name: str):
        self.record_method('ReferenceCoordSystem', name)

    def set_coord_system_type(self, name: str):
        self.record_method('CoordSystemType', name)

    def set_corrugation(self, depth: float, gap_width: float, tooth_width: float):
        self.record_method('Corrugation', depth, gap_width, tooth_width)

    def set_ohmic_sheet_impedance(self, impedance: complex):
        self.record_method('OhmicSheetImpedance', impedance.real, impedance.imag)

    def set_ohmic_sheet_frequency(self, frequency: float):
        self.record_method('OhmicSheetFreq', frequency)

    def set_tabulated_surface_impedance_model(self, model: str):
        self.record_method('SetTabulatedSurfaceImpedanceModel', model)

    def add_tabulated_surface_impedance_fitting_value(
            self, frequency: float, impedance: complex, weight: float):
        self.record_method(
            'AddTabulatedSurfaceImpedanceFittingValue', frequency,
            impedance.real, impedance.imag, weight)

    def add_tabulated_surface_impedance_value(self, frequency: float, impedance: complex):
        self.record_method(
            'AddTabulatedSurfaceImpedanceValue', frequency, impedance.real, impedance.imag)

    def set_dispersive_fitting_scheme_tab_surf_imp(self, scheme: str):
        self.record_method('DispersiveFittingSchemeTabSI', scheme)

    def set_max_order_nth_model_fit_tab_surf_imp(self, order: int):
        self.record_method('MaximalOrderNthModelFitTabSI', order)

    def set_use_only_data_in_sim_freq_range_nth_model_tab_surf_imp(self, flag: bool = True):
        self.record_method('UseOnlyDataInSimFreqRangeNthModelTabSI', flag)

    def set_error_limit_nth_model_fit_tab_surf_imp(self, limit: float):
        self.record_method('ErrorLimitNthModelFitTabSI', limit)

    def set_as_tabulated_compact_model(self, flag: bool = True):
        self.record_method('TabulatedCompactModel', flag)

    def reset_tabulated_compact_model_list(self):
        self.record_method('ResetTabulatedCompactModelList')

    def set_tabulated_compact_model_impedance(self, imp_p1: float, imp_p2: float):
        self.record_method('SetTabulatedCompactModelImpedance', imp_p1, imp_p2)

    def set_symm_tabulated_compact_model_impedance(self, imp_p: float):
        self.record_method('SetSymmTabulatedCompactModelImpedance', imp_p)

    def add_tabulated_compact_model_item(
            self, frequency: float, s11: complex, s21: complex, s12: complex,
            s22: complex, weight: float):
        self.record_method(
            'AddTabulatedCompactModelItem', frequency, s11.real, s11.imag, s21.real, s21.imag,
            s12.real, s12.imag, s22.real, s22.imag, weight)

    def add_symm_tabulated_compact_model_item(
            self, frequency: float, s11: complex, s21: complex, weight: float):
        self.record_method(
            'AddSymmTabulatedCompactModelItem', frequency, s11.real, s11.imag,
            s21.real, s21.imag, weight)

    def set_tabulated_compact_model_anisotropic(self, flag: bool = True):
        self.record_method('TabulatedCompactModelAnisotropic', flag)

    def add_aniso_tabulated_compact_model_item(
            self, entry_block: str, frequency: float, s11: complex, s21: complex, s12: complex,
            s22: complex, weight: float):
        self.record_method(
            'AddAnisoTabulatedCompactModelItem', entry_block, frequency,
            s11.real, s11.imag, s21.real, s21.imag,
            s12.real, s12.imag, s22.real, s22.imag, weight)

    def add_aniso_symm_tabulated_compact_model_item(
            self, entry_block: str, frequency: float, s11: complex, s21: complex, weight: float):
        self.record_method(
            'AddAnisoSymmTabulatedCompactModelItem', entry_block, frequency,
            s11.real, s11.imag, s21.real, s21.imag, weight)

    def set_max_order_fit_tabulated_compact_model(self, order: int):
        self.record_method('MaximalOrderFitTabulatedCompactModel', order)

    def set_use_only_data_in_sim_freq_range_tabulated_compact_model(self, flag: bool = True):
        self.record_method('UseOnlyDataInSimFreqRangeTabulatedCompactModel', flag)

    def set_error_limit_fit_tabulated_compact_model(self, limit: float):
        self.record_method('ErrorLimitFitTabulatedCompactModel', limit)

    def set_thickness(self, thickness: float):
        self.record_method('Thickness', thickness)

    def set_lossy_metal_surf_imp_roughness(self, roughness: float):
        self.record_method('LossyMetalSIRoughness', roughness)

    def set_el_conductivity(self, sigma: float):
        self.record_method('Sigma', sigma)

    def set_el_conductivity_x(self, sigma: float):
        self.record_method('SigmaX', sigma)

    def set_el_conductivity_y(self, sigma: float):
        self.record_method('SigmaY', sigma)

    def set_el_conductivity_z(self, sigma: float):
        self.record_method('SigmaZ', sigma)

    def set_el_parametric_conductivity(self, flag: bool = True):
        self.record_method('SetElParametricConductivity', flag)

    def add_je_value(self, dj: float, de: float):
        self.record_method('AddJEValue', dj, de)

    def reset_je_list(self):
        self.record_method('ResetJEList')

    def reset_el_time_dependent_conductivity_curve(self):
        self.record_method('ResetElTimeDepCond')

    def add_el_time_dependent_conductivity_value(self, d_time: float, d_cond: float):
        self.record_method('AddElTimeDepCondValue', d_time, d_cond)

    def add_el_time_dependent_conductivity_aniso_value(
            self, d_time: float, d_cond_x: float, d_cond_y: float, d_cond_z: float):
        self.record_method('AddElTimeDepCondAnisoValue', d_time, d_cond_x, d_cond_y, d_cond_z)

    def load_el_time_dependent_conductivity_from_file(self, file_path: str, time_unit: str):
        self.record_method('LoadElTimeDepConductivity', file_path, time_unit)

    def set_el_tangent_delta(self, tan_d: float):
        self.record_method('TanD', tan_d)

    def set_el_tangent_delta_x(self, tan_d: float):
        self.record_method('TanDX', tan_d)

    def set_el_tangent_delta_y(self, tan_d: float):
        self.record_method('TanDY', tan_d)

    def set_el_tangent_delta_z(self, tan_d: float):
        self.record_method('TanDZ', tan_d)

    def set_el_tangent_delta_given(self, flag: bool = True):
        self.record_method('TanDGiven', flag)

    def set_el_tangent_delta_frequency(self, frequency: float):
        self.record_method('TanDFreq', frequency)

    def set_el_tangent_delta_model(self, model: str):
        self.record_method('TanDModel', model)

    def set_el_const_tangent_delta_strategy_eps(self, strategy: str):
        self.record_method('SetConstTanDStrategyEps', strategy)

    def set_el_const_tangent_delta_model_order_eps(self, order: int):
        self.record_method('ConstTanDModelOrderEps', order)

    def set_djordjevic_sarkar_upper_freq_eps(self, frequency: float):
        self.record_method('DjordjevicSarkarUpperFreqEps', frequency)

    def set_mag_conductivity(self, sigma: float):
        self.record_method('SigmaM', sigma)

    def set_mag_conductivity_x(self, sigma: float):
        self.record_method('SigmaMX', sigma)

    def set_mag_conductivity_y(self, sigma: float):
        self.record_method('SigmaMY', sigma)

    def set_mag_conductivity_z(self, sigma: float):
        self.record_method('SigmaMZ', sigma)

    def set_mag_parametric_conductivity(self, flag: bool = True):
        self.record_method('SetMagParametricConductivity', flag)

    def reset_mag_time_dependent_conductivity_curve(self):
        self.record_method('ResetMagTimeDepCond')

    def add_mag_time_dependent_conductivity_value(self, d_time: float, d_cond: float):
        self.record_method('AddMagTimeDepCondValue', d_time, d_cond)

    def add_mag_time_dependent_conductivity_aniso_value(
            self, d_time: float, d_cond_x: float, d_cond_y: float, d_cond_z: float):
        self.record_method('AddMagTimeDepCondAnisoValue', d_time, d_cond_x, d_cond_y, d_cond_z)

    def load_mag_time_dependent_conductivity_from_file(self, file_path: str, time_unit: str):
        self.record_method('LoadMagTimeDepConductivity', file_path, time_unit)

    def set_mag_tangent_delta(self, tan_d: float):
        self.record_method('TanDM', tan_d)

    def set_mag_tangent_delta_x(self, tan_d: float):
        self.record_method('TanDMX', tan_d)

    def set_mag_tangent_delta_y(self, tan_d: float):
        self.record_method('TanDMY', tan_d)

    def set_mag_tangent_delta_z(self, tan_d: float):
        self.record_method('TanDMZ', tan_d)

    def set_mag_tangent_delta_given(self, flag: bool = True):
        self.record_method('TanDMGiven', flag)

    def set_mag_tangent_delta_frequency(self, frequency: float):
        self.record_method('TanDMFreq', frequency)

    def set_mag_tangent_delta_model(self, model: str):
        self.record_method('TanDMModel', model)

    def set_mag_const_tangent_delta_strategy_mu(self, strategy: str):
        self.record_method('SetConstTanDStrategyMu', strategy)

    def set_mag_const_tangent_delta_model_order_mu(self, order: int):
        self.record_method('ConstTanDModelOrderMu', order)

    def set_djordjevic_sarkar_upper_freq_mu(self, frequency: float):
        self.record_method('DjordjevicSarkarUpperFreqMu', frequency)

    def set_disp_model_esp(self, model: str):
        self.record_method('DispModelEps', model)

    def set_disp_model_mu(self, model: str):
        self.record_method('DispModelMu', model)

    def set_eps_infinity(self, eps: float):
        self.record_method('EpsInfinity', eps)

    def set_eps_infinity_x(self, eps: float):
        self.record_method('EpsInfinityX', eps)

    def set_eps_infinity_y(self, eps: float):
        self.record_method('EpsInfinityY', eps)

    def set_eps_infinity_z(self, eps: float):
        self.record_method('EpsInfinityZ', eps)

    def set_disp_coeff_eps(self, index: int, eps: float):
        if index < 0 or index > 4:
            raise ValueError('Invalid index.')
        self.record_method('DispCoeff'+str(index)+'Eps', eps)

    def set_disp_coeff_eps_x(self, index: int, eps: float):
        if index < 0 or index > 4:
            raise ValueError('Invalid index.')
        self.record_method('DispCoeff'+str(index)+'EpsX', eps)

    def set_disp_coeff_eps_y(self, index: int, eps: float):
        if index < 0 or index > 4:
            raise ValueError('Invalid index.')
        self.record_method('DispCoeff'+str(index)+'EpsY', eps)

    def set_disp_coeff_eps_z(self, index: int, eps: float):
        if index < 0 or index > 4:
            raise ValueError('Invalid index.')
        self.record_method('DispCoeff'+str(index)+'EpsZ', eps)

    def add_disp_eps_pole_1st_order(self, alpha0: float, beta0: float):
        self.record_method('AddDispEpsPole1stOrder', alpha0, beta0)

    def add_disp_eps_pole_1st_order_x(self, alpha0: float, beta0: float):
        self.record_method('AddDispEpsPole1stOrderX', alpha0, beta0)

    def add_disp_eps_pole_1st_order_y(self, alpha0: float, beta0: float):
        self.record_method('AddDispEpsPole1stOrderY', alpha0, beta0)

    def add_disp_eps_pole_1st_order_z(self, alpha0: float, beta0: float):
        self.record_method('AddDispEpsPole1stOrderZ', alpha0, beta0)

    def add_disp_eps_pole_2nd_order(
            self, alpha0: float, alpha1: float, beta0: float, beta1: float):
        self.record_method('AddDispEpsPole2ndOrder', alpha0, alpha1, beta0, beta1)

    def add_disp_eps_pole_2nd_order_x(
            self, alpha0: float, alpha1: float, beta0: float, beta1: float):
        self.record_method('AddDispEpsPole2ndOrderX', alpha0, alpha1, beta0, beta1)

    def add_disp_eps_pole_2nd_order_y(
            self, alpha0: float, alpha1: float, beta0: float, beta1: float):
        self.record_method('AddDispEpsPole2ndOrderY', alpha0, alpha1, beta0, beta1)

    def add_disp_eps_pole_2nd_order_z(
            self, alpha0: float, alpha1: float, beta0: float, beta1: float):
        self.record_method('AddDispEpsPole2ndOrderZ', alpha0, alpha1, beta0, beta1)

    def add_disp_mu_pole_1st_order(self, alpha0: float, beta0: float):
        self.record_method('AddDispMuPole1stOrder', alpha0, beta0)

    def add_disp_mu_pole_1st_order_x(self, alpha0: float, beta0: float):
        self.record_method('AddDispMuPole1stOrderX', alpha0, beta0)

    def add_disp_mu_pole_1st_order_y(self, alpha0: float, beta0: float):
        self.record_method('AddDispMuPole1stOrderY', alpha0, beta0)

    def add_disp_mu_pole_1st_order_z(self, alpha0: float, beta0: float):
        self.record_method('AddDispMuPole1stOrderZ', alpha0, beta0)

    def add_disp_mu_pole_2nd_order(
            self, alpha0: float, alpha1: float, beta0: float, beta1: float):
        self.record_method('AddDispMuPole2ndOrder', alpha0, alpha1, beta0, beta1)

    def add_disp_mu_pole_2nd_order_x(
            self, alpha0: float, alpha1: float, beta0: float, beta1: float):
        self.record_method('AddDispMuPole2ndOrderX', alpha0, alpha1, beta0, beta1)

    def add_disp_mu_pole_2nd_order_y(
            self, alpha0: float, alpha1: float, beta0: float, beta1: float):
        self.record_method('AddDispMuPole2ndOrderY', alpha0, alpha1, beta0, beta1)

    def add_disp_mu_pole_2nd_order_z(
            self, alpha0: float, alpha1: float, beta0: float, beta1: float):
        self.record_method('AddDispMuPole2ndOrderZ', alpha0, alpha1, beta0, beta1)

    def set_mu_infinity(self, mu: float):
        self.record_method('MuInfinity', mu)

    def set_mu_infinity_x(self, mu: float):
        self.record_method('MuInfinityX', mu)

    def set_mu_infinity_y(self, mu: float):
        self.record_method('MuInfinityY', mu)

    def set_mu_infinity_z(self, mu: float):
        self.record_method('MuInfinityZ', mu)

    def set_disp_coeff_mu(self, index: int, mu: float):
        if index < 0 or index > 4:
            raise ValueError('Invalid index.')
        self.record_method('DispCoeff'+str(index)+'Mu', mu)

    def set_disp_coeff_mu_x(self, index: int, mu: float):
        if index < 0 or index > 4:
            raise ValueError('Invalid index.')
        self.record_method('DispCoeff'+str(index)+'MuX', mu)

    def set_disp_coeff_mu_y(self, index: int, mu: float):
        if index < 0 or index > 4:
            raise ValueError('Invalid index.')
        self.record_method('DispCoeff'+str(index)+'MuY', mu)

    def set_disp_coeff_mu_z(self, index: int, mu: float):
        if index < 0 or index > 4:
            raise ValueError('Invalid index.')
        self.record_method('DispCoeff'+str(index)+'MuZ', mu)

    def set_use_si_unit_system(self, flag: bool = True):
        self.record_method('UseSISystem', flag)

    def set_gyro_mu_frequency(self, frequency: float):
        self.record_method('GyroMuFreq', frequency)

    def set_magnetostatic_dep_source_field(self, source_field: str):
        self.record_method('SetMagnetostaticDepSourceField', source_field)

    def set_dispersive_fitting_format_eps(self, format: str):
        self.record_method('DispersiveFittingFormatEps', format)

    def set_dispersive_fitting_format_mu(self, format: str):
        self.record_method('DispersiveFittingFormatMu', format)

    def add_dispersion_fitting_value_eps(self, d_freq: float, d_val: complex, d_weight: float):
        self.record_method('AddDispersionFittingValueEps', d_freq, d_val.real, d_val.imag, d_weight)

    def add_dispersion_fitting_value_mu(self, d_freq: float, d_val: complex, d_weight: float):
        self.record_method('AddDispersionFittingValueMu', d_freq, d_val.real, d_val.imag, d_weight)

    def add_dispersion_fitting_value_xyz_eps(
            self, d_freq: float, d_val_x: complex, d_val_y: complex,
            d_val_z: complex, d_weight: float):
        self.record_method(
            'AddDispersionFittingValueXYZEps', d_freq, d_val_x, d_val_y, d_val_z, d_weight)

    def add_dispersion_fitting_value_xyz_mu(
            self, d_freq: float, d_val_x: complex, d_val_y: complex,
            d_val_z: complex, d_weight: float):
        self.record_method(
            'AddDispersionFittingValueXYZMu', d_freq, d_val_x, d_val_y, d_val_z, d_weight)

    def set_dispersive_fitting_scheme_eps(self, scheme: str):
        self.record_method('DispersiveFittingSchemeEps', scheme)

    def set_dispersive_fitting_scheme_mu(self, scheme: str):
        self.record_method('DispersiveFittingSchemeMu', scheme)

    def set_max_order_nth_model_fit_eps(self, order: int):
        self.record_method('MaximalOrderNthModelFitEps', order)

    def set_max_order_nth_model_fit_mu(self, order: int):
        self.record_method('MaximalOrderNthModelFitMu', order)

    def set_error_limit_nth_model_fit_esp(self, limit: float):
        # not documented, found in the history list
        self.record_method('ErrorLimitNthModelFitEps', limit)

    def set_error_limit_nth_model_fit_mu(self, limit: float):
        # not documented, found in the history list
        self.record_method('ErrorLimitNthModelFitMu', limit)

    def set_use_only_data_in_sim_freq_range_nth_model_eps(self, flag: bool = True):
        self.record_method('UseOnlyDataInSimFreqRangeNthModelEps', flag)

    def set_use_only_data_in_sim_freq_range_nth_model_mu(self, flag: bool = True):
        self.record_method('UseOnlyDataInSimFreqRangeNthModelMu', flag)

    def set_use_general_dispersion_eps(self, flag: bool = True):
        self.record_method('UseGeneralDispersionEps', flag)

    def set_use_general_dispersion_mu(self, flag: bool = True):
        self.record_method('UseGeneralDispersionMu', flag)

    def set_tensor_formula_for_eps_r(self):
        self.record_method('TensorFormulaFor', 'epsilon_r')

    def set_tensor_formula_for_mu_r(self):
        self.record_method('TensorFormulaFor', 'mu_r')

    def set_tensor_formula_real(self, row: int, column: int, formula: str):
        self.record_method('TensorFormulaReal', row, column, formula)

    def set_tensor_formula_imag(self, row: int, column: int, formula: str):
        self.record_method('TensorFormulaImag', row, column, formula)

    def set_tensor_alignment(self, w_x: float, w_y: float, w_z: float):
        self.record_method('TensorAlignment', w_x, w_y, w_z)

    def set_tensor_alignment2(
            self, u_x: float, u_y: float, u_z: float, v_x: float, v_y: float, v_z: float):
        self.record_method('TensorAlignment2', u_x, u_y, u_z, v_x, v_y, v_z)

    def reset_spatially_varying_material_parameter(self, prop: str):
        self.record_method('ResetSpatiallyVaryingMaterialParameter', prop)

    def set_spatially_varying_material_model(self, prop: str, model: str):
        self.record_method('SpatiallyVaryingMaterialModel', prop, model)

    def set_spatially_varying_material_model_aniso(self, prop: str, direction: str, model: str):
        self.record_method('SpatiallyVaryingMaterialModelAniso', prop, direction, model)

    def add_spatially_varying_material_parameter(self, prop: str, param: str, value: float):
        self.record_method('AddSpatiallyVaryingMaterialParameter', prop, param, value)

    def add_spatially_varying_material_parameter_aniso(
            self, prop: str, direction: str, param: str, value: float):
        self.record_method(
            'AddSpatiallyVaryingMaterialParameterAniso', prop, direction, param, value)

    def reset_space_map_material(self, prop: str):
        self.record_method('ResetSpaceMapBasedMaterial', prop)

    def set_space_map_model(self, prop: str, model: str):
        self.record_method('SpaceMapBasedOperator', prop, model)

    def set_space_map_model_aniso(self, prop: str, direction: str, model: str):
        self.record_method('SpaceMapBasedOperatorAniso', prop, direction, model)

    def add_space_map_material_parameter(
            self, prop: str, param: str, value: Union[str, float, int]):
        if isinstance(value, str):
            self.record_method('AddSpaceMapBasedMaterialStringParameter', prop, param, value)
        else:
            self.record_method('AddSpaceMapBasedMaterialDoubleParameter', prop, param, value)

    def add_space_map_material_parameter_aniso(
            self, prop: str, direction: str, param: str, value: Union[str, float, int]):
        if isinstance(value, str):
            self.record_method(
                'AddSpaceMapBasedMaterialStringParameterAniso', prop, direction, param, value)
        else:
            self.record_method(
                'AddSpaceMapBasedMaterialDoubleParameterAniso', prop, direction, param, value)

    def convert_material_field(self, file_path_in: str, file_path_out: str):
        self.record_method('ConvertMaterialField', file_path_in, file_path_out)

    def reset_tab_mu_differential(self):
        self.record_method('ResetTabMuDifferential')

    def add_tab_mu_differential(self, h_field: float, mu_diff: float):
        self.record_method('AddTabMuDifferential', h_field, mu_diff)

    def set_generalized_debye_non_lin_dependency(self, dependency: str):
        self.record_method('GeneralizedDebyeNonLinDependency', dependency)

    def set_magnetostatic_dep_source_field(self, source_field: str):
        self.record_method('SetMagnetostaticDepSourceField', source_field)

    def set_coating_type_definition(self, prop: str):
        self.record_method('SetCoatingTypeDefinition', prop)

    def add_tabulated_surface_impedance_deg(
            self, frequency: float, angle: float, z_te: complex, z_tm: complex):
        self.record_method(
            'AddTabulatedSurfaceImpedance', frequency, angle,
            z_te.real, z_te.imag, z_tm.real, z_tm.imag)

    def add_tabulated_surface_impedance_rad(
            self, frequency: float, angle: float, z_te: complex, z_tm: complex):
        self.add_tabulated_surface_impedance_deg(
            frequency, np.rad2deg(angle), z_te.real, z_te.imag, z_tm.real, z_tm.imag)

    def add_tabulated_reflection_factor_deg(
            self, frequency: float, angle: float, r_te: complex, r_tm: complex):
        self.record_method(
            'AddTabulatedReflectionFactor', frequency, angle,
            r_te.real, r_te.imag, r_tm.real, r_tm.imag)

    def add_tabulated_reflection_factor_rad(
            self, frequency: float, angle: float, r_te: complex, r_tm: complex):
        self.add_tabulated_reflection_factor_deg(
            frequency, np.rad2deg(angle), r_te.real, r_te.imag, r_tm.real, r_tm.imag)

    def add_tabulated_reflection_transmission_factor_deg(
            self, frequency: float, angle: float,
            r_te: complex, r_tm: complex, t_te: complex, t_tm: complex):
        self.record_method(
            'AddTabulatedReflectionTransmissionFactor', frequency, angle,
            r_te.real, r_te.imag, r_tm.real, r_tm.imag, t_te.real, t_te.imag, t_tm.real, t_tm.imag)

    def add_tabulated_reflection_transmission_factor_rad(
            self, frequency: float, angle: float,
            r_te: complex, r_tm: complex, t_te: complex, t_tm: complex):
        self.add_tabulated_reflection_transmission_factor_deg(
            frequency, np.rad2deg(angle),
            r_te.real, r_te.imag, r_tm.real, r_tm.imag, t_te.real, t_te.imag, t_tm.real, t_tm.imag)

    def add_temperature_dep_eps(self, d_temperature: float, d_value: float):
        self.record_method('AddTemperatureDepEps', d_temperature, d_value)

    def reset_temperature_dep_eps(self):
        self.record_method('ResetTemperatureDepEps')

    def add_temperature_dep_mu(self, d_temperature: float, d_value: float):
        self.record_method('AddTemperatureDepMu', d_temperature, d_value)

    def reset_temperature_dep_mu(self):
        self.record_method('ResetTemperatureDepMu')

    def add_temperature_dep_sigma(self, d_temperature: float, d_value: float):
        self.record_method('AddTemperatureDepSigma', d_temperature, d_value)

    def reset_temperature_dep_sigma(self):
        self.record_method('ResetTemperatureDepSigma')

    def set_temperature_dep_source_field(self, field_name: str):
        self.record_method('SetTemperatureDepSourceField', field_name)

    def add_hb_value(self, h_value: float, b_value: float):
        self.record_method('AddHBValue', h_value, b_value)

    def reset_hb_list(self):
        self.record_method('ResetHBList')

    def set_use_nl_anisotropy(self, flag: bool = True):
        self.record_method('NLAnisotropy', flag)

    def set_nla_stacking_factor(self, factor: float):
        self.record_method('NLAStackingFactor', factor)

    def set_nla_direction_x(self, dir: float):
        self.record_method('NLADirectionX', dir)

    def set_nla_direction_y(self, dir: float):
        self.record_method('NLADirectionY', dir)

    def set_nla_direction_z(self, dir: float):
        self.record_method('NLADirectionZ', dir)

    def set_nla_direction(self, x: float, y: float, z: float):
        self.set_nla_direction_x(x)
        self.set_nla_direction_y(y)
        self.set_nla_direction_z(z)

    def set_particle_property(self, prop: str):
        self.record_method('ParticleProperty', prop)

    def set_se_mdoel(self, model: str):
        self.record_method('SeModel', model)

    def set_se_max_number_of_generations(self, number: int):
        self.record_method('SeMaxGenerations', number)

    def set_se_max_number_of_secondaries(self, number: int):
        self.record_method('SeMaxSecondaries', number)

    def set_se_ts_param_t1(self, value: float):
        self.record_method('SeTsParam_T1', value)

    def set_se_ts_param_t2(self, value: float):
        self.record_method('SeTsParam_T2', value)

    def set_se_ts_param_t3(self, value: float):
        self.record_method('SeTsParam_T3', value)

    def set_se_ts_param_t4(self, value: float):
        self.record_method('SeTsParam_T4', value)

    def set_se_ts_param_sey(self, value: float):
        self.record_method('SeTsParam_SEY', value)

    def set_se_ts_param_energy(self, value: float):
        self.record_method('SeTsParam_Energy', value)

    def set_se_ts_param_s(self, value: float):
        self.record_method('SeTsParam_S', value)

    def set_se_ts_param_pn(self, n: int, value: float):
        self.record_method('SeTsParam_PN', n, value)

    def set_se_ts_param_epsn(self, n: int, value: float):
        self.record_method('SeTsParam_EpsN', n, value)

    def set_se_rd_param_r(self, value: float):
        self.record_method('SeRdParam_R', value)

    def set_se_rd_param_r1(self, value: float):
        self.record_method('SeRdParam_R1', value)

    def set_se_rd_param_r2(self, value: float):
        self.record_method('SeRdParam_R2', value)

    def set_se_rd_param_q(self, value: float):
        self.record_method('SeRdParam_Q', value)

    def set_se_rd_param_p1inf(self, value: float):
        self.record_method('SeRdParam_P1Inf', value)

    def set_se_rd_param_energy(self, value: float):
        self.record_method('SeRdParam_Energy', value)

    def set_se_bs_param_sigma(self, value: float):
        self.record_method('SeBsParam_Sigma', value)

    def set_se_bs_param_e1(self, value: float):
        self.record_method('SeBsParam_E1', value)

    def set_se_bs_param_e2(self, value: float):
        self.record_method('SeBsParam_E2', value)

    def set_se_bs_param_p1hat(self, value: float):
        self.record_method('SeBsParam_P1Hat', value)

    def set_se_bs_param_p1inf(self, value: float):
        self.record_method('SeBsParam_P1Inf', value)

    def set_se_bs_param_energy(self, value: float):
        self.record_method('SeBsParam_Energy', value)

    def set_se_bs_param_p(self, value: float):
        self.record_method('SeBsParam_P', value)

    def set_se_bs_param_w(self, value: float):
        self.record_method('SeBsParam_W', value)

    def enable_se_plot_1d_deg(self, angle: float, energy: float):
        self.record_method('SePlot1D', True, angle, energy)

    def enable_se_plot_1d_rad(self, angle: float, energy: float):
        self.enable_se_plot_1d_deg(np.rad2deg(angle), energy)

    def disable_se_plot_1d_deg(self):
        self.record_method('SePlot1D', False, 0, 0)

    def setup_se_vaughan(
            self, energy_max: float, se_yield_max: float, energy_threshold: float,
            smoothness: float, temperature: float):
        self.record_method(
            'SeVaughan', energy_max, se_yield_max, energy_threshold, smoothness, temperature)

    def set_se_import_settings(self, file_name: str, temperature: float):
        self.record_method('SeImportSettings', file_name, temperature)

    def set_se_import_data(self, energy: float, sey: float):
        self.record_method('SeImportData', energy, sey)

    def set_ion_see_model(self, model: str):
        self.record_method('IonSEEModel', model)

    def set_ion_see_import_settings(self, file_name: str, temperature: float):
        self.record_method('IonSEEImportSettings', file_name, temperature)

    def set_ion_see_import_data(self, energy: float, sey: float):
        self.record_method('IonSEEImportData', energy, sey)

    def set_particle_transparency_settings(self, type: str, percent: float, file_name: str):
        self.record_method('ParticleTransparencySettings', type, percent, file_name)

    def set_particle_transparency_import_data(self, energy: float, transparency: float):
        self.record_method('ParticleTransparencyImportData', energy, transparency)

    def set_energy_step_for_sey_plots(self, step: float):
        self.record_method('SetEnergyStepForSEYPlots', step)

    def set_special_disp_param_for_pic(
            self, esp_inf: float, relax_freq: float, reson_freq: float, reson_width: float,
            lorentz_weight: float):
        self.record_method(
            'SpecialDispParamForPIC', esp_inf, relax_freq, reson_freq, reson_width, lorentz_weight)

    def set_special_disp_param_visual(self, max_freq: float, prop_dist: float):
        self.record_method('SpecialDispParamVisual', max_freq, prop_dist)

    def set_thermal_type(self, thermal_type: str):
        self.record_method('ThermalType', thermal_type)

    def set_thermal_conductivity(self, value: float):
        self.record_method('ThermalConductivity', value)

    def set_thermal_conductivity_x(self, value: float):
        self.record_method('ThermalConductivityX', value)

    def set_thermal_conductivity_y(self, value: float):
        self.record_method('ThermalConductivityY', value)

    def set_thermal_conductivity_z(self, value: float):
        self.record_method('ThermalConductivityZ', value)

    def set_specific_heat(self, value: float, unit: str):
        self.record_method('SpecificHeat', value, unit)

    def set_blood_flow(self, value: float):
        self.record_method('BloodFlow', value)

    def set_metabolic_rate(self, value: float):
        self.record_method('MetabolicRate', value)

    def set_voxel_convection(self, value: float):
        self.record_method('VoxelConvection', value)

    def set_dynamic_viscosity(self, value: float):
        self.record_method('DynamicViscosity', value)

    def set_emissivity(self, value: float):
        self.record_method('Emissivity', value)

    def reset_nl_thermal_cond(self):
        self.record_method('ResetNLThermalCond')

    def add_nl_thermal_cond(self, temperature: float, value: float):
        self.record_method('AddNLThermalCond', temperature, value)

    def add_nl_thermal_cond_aniso(
            self, temperature: float, value_x: float, value_y: float, value_z: float):
        self.record_method('AddNLThermalCondAniso', temperature, value_x, value_y, value_z)

    def reset_nl_heat_cap(self):
        self.record_method('ResetNLHeatCap')

    def add_nl_specific_heat(self, temperature: float, value: float, unit: str):
        self.record_method('AddNLSpecificHeat', temperature, value, unit)

    def reset_nl_blook_flow(self):
        self.record_method('ResetNLBloodFlow')

    def add_nl_blood_flow_min_temperature(self, value: float):
        self.record_method('SetNLBloodFlowMinTemperature', value)

    def add_nl_blood_flow_basal_value(self, value: float):
        self.record_method('SetNLBloodFlowBasalValue', value)

    def add_nl_blood_flow_local_vasodilation_param(self, value: float):
        self.record_method('SetNLBloodFlowLocalVasodilationParam', value)

    def add_nl_blood_flow_max_multiplier(self, value: float):
        self.record_method('SetNLBloodFlowMaxMultiplier', value)

    def add_nl_dynamic_viscosity(self, temperature: float, value: float):
        self.record_method('AddNLDynamicViscosity', temperature, value)

    def set_mechanics_type(self, mechanics_type: str):
        self.record_method('MechanicsType', mechanics_type)

    def set_youngs_modulus(self, value: float):
        self.record_method('YoungsModulus', value)

    def set_poisson_ratio(self, value: float):
        self.record_method('PoissonsRatio', value)

    def set_thermal_expansion_rate(self, value: float):
        self.record_method('ThermalExpansionRate', value)

    def reset_temp_dep_youngs_modulus(self):
        self.record_method('ResetTempDepYoungsModulus')

    def add_temp_dep_youngs_modulus(self, temperature: float, value: float):
        self.record_method('AddTempDepYoungsModulus', temperature, value)

    def set_intrinsic_carrier_density(self, value: float):
        self.record_method('IntrinsicCarrierDensity', value)

    def set_lattice_scattering(self, species: str, mobility: float, exponent: float):
        self.record_method('LatticeScattering', species, mobility, exponent)

    def set_effective_mass_for_conductivity(self, species: str, effective_mass: float):
        self.record_method('EffectiveMassForConductivity', species, effective_mass)

    def enable_auger_recombination(self, gamma_electron: float, gamma_hole: float):
        self.record_method('AugerRecombination', True, gamma_electron, gamma_hole)

    def disable_auger_recombination(self):
        self.record_method('AugerRecombination', False, 0, 0)

    def enable_band_to_band_recombination(self, rate: float):
        self.record_method('BandToBandRecombination', True, rate)

    def disable_band_to_band_recombination(self):
        self.record_method('BandToBandRecombination', False, 0)

    def enable_impact_ionization(
            self, electron_ionization_rate: float, electron_exponent: float,
            electron_critical_field: float, hole_ionization_rate: float, hole_exponent: float,
            hole_critical_field: float):
        self.record_method(
            'ImpactIonization', True, electron_ionization_rate, electron_exponent,
            electron_critical_field, hole_ionization_rate, hole_exponent, hole_critical_field)

    def disable_impact_ionization(self):
        self.record_method('ImpactIonization', False, 0, 0, 0, 0, 0, 0)

    def enable_optically_induced_carrier_generation(self, quantum_efficiency: float):
        self.record_method('OpticallyInducedCarrierGeneration', True, quantum_efficiency)

    def disable_optically_induced_carrier_generation(self):
        self.record_method('OpticallyInducedCarrierGeneration', False, 0)

    def enable_srh_recombination(
            self, trap_energy_level: float, electron_lifetime: float,
            electron_reference_density: float, hole_lifetime: float, hole_reference_density: float):
        self.record_method(
            'SRHRecombination', True, trap_energy_level, electron_lifetime,
            electron_reference_density, hole_lifetime, hole_reference_density)

    def disable_srh_recombination(self):
        self.record_method('SRHRecombination', False, 0, 0, 0, 0, 0)

    def set_flow_res_pressure_loss_type_u(self, value: str):
        self.record_method('FlowResPressureLossTypeU', value)

    def set_flow_res_pressure_loss_type_v(self, value: str):
        self.record_method('FlowResPressureLossTypeV', value)

    def set_flow_res_pressure_loss_type_w(self, value: str):
        self.record_method('FlowResPressureLossTypeW', value)

    def set_flow_res_loss_coefficient_u(self, value: float):
        self.record_method('FlowResLossCoefficientU', value)

    def set_flow_res_loss_coefficient_v(self, value: float):
        self.record_method('FlowResLossCoefficientV', value)

    def set_flow_res_loss_coefficient_w(self, value: float):
        self.record_method('FlowResLossCoefficientW', value)

    def set_flow_res_pressure_loss_type_sheet(self, value: str):
        self.record_method('FlowResPressureLossTypeSheet', value)

    def set_flow_res_loss_coefficient_sheet(self, value: float):
        self.record_method('FlowResLossCoefficientSheet', value)

    def set_flow_res_free_area_ratio(self, value: float):
        self.record_method('FlowResFreeAreaRatio', value)

    def set_flow_res_shape_type(self, value: str):
        self.record_method('FlowResShapeType', value)

    def set_flow_res_shape_size(self, value: float):
        self.record_method('FlowResShapeSize', value)

    def set_flow_res_shape_u_pitch(self, value: float):
        self.record_method('FlowResShapeUPitch', value)

    def set_flow_res_shape_v_pitch(self, value: float):
        self.record_method('FlowResShapeVPitch', value)

    def get_number_of_materials(self) -> int:
        return self.record_method('GetNumberOfMaterials')

    def get_name_of_material_from_index(self) -> str:
        return self.record_method('GetNameOfMaterialFromIndex')

    def is_background_material(self, material_name: str) -> bool:
        return self.record_method('IsBackgroundMaterial', material_name)

    def get_type_of_background_material(self) -> str:
        return self.record_method('GetTypeOfBackgroundMaterial')

    def get_type_of_material(self, material_name: str) -> str:
        return self.record_method('GetTypeOfMaterial', material_name)

    def get_color_rgb(self, material_name: str) -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetColour', B, material_name, D, D, D)
        return None if not res[0] else res[1:]

    # returns: (eps_x, eps_y, eps_z)
    def get_epsilon(self, material_name: str) -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetEpsilon', B, material_name, D, D, D)
        return None if not res[0] else res[1:]

    # returns: (mu_x, mu_y, mu_z)
    def get_mu(self, material_name: str) -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetMu', B, material_name, D, D, D)
        return None if not res[0] else res[1:]

    # returns: (sigma_x, sigma_y, sigma_z)
    def get_el_sigma(self, material_name: str) -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetSigma', B, material_name, D, D, D)
        return None if not res[0] else res[1:]

    # returns: (sigma_x, sigma_y, sigma_z)
    def get_mag_sigma(self, material_name: str) -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetSigmaM', B, material_name, D, D, D)
        return None if not res[0] else res[1:]

    # returns: (depth, gap_width, tooth_width)
    def get_corrugation(self, material_name: str) -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetCorrugation', B, material_name, D, D, D)
        return None if not res[0] else res[1:]

    def get_ohmic_sheet_impedance(self, material_name: str) -> Optional[complex]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetOhmicSheetImpedance', B, material_name, D, D)
        return None if not res[0] else complex(res[1], res[2])

    def get_rho(self, material_name: str) -> Optional[float]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetRho', B, material_name, D)
        return None if not res[0] else res[1]

    def get_dynamic_viscosity(self, material_name: str) -> Optional[float]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetDynamicViscosity', B, material_name, D)
        return None if not res[0] else res[1]

    def get_dynamic_viscosity(self, material_name: str) -> Optional[float]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetEmissivity', B, material_name, D)
        return None if not res[0] else res[1]

    # returns: (thermal_conductivity_x, thermal_conductivity_y, thermal_conductivity_z)
    def get_thermal_conductivity(self, material_name: str) -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetThermalConductivity', B, material_name, D, D, D)
        return None if not res[0] else res[1:]

    def get_specific_heat(self, material_name: str) -> Optional[float]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetSpecificHeat', B, material_name, D)
        return None if not res[0] else res[1]

    def get_specific_heat(self, material_name: str) -> Optional[float]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetHeatCapacity', B, material_name, D)
        return None if not res[0] else res[1]

    def get_blood_flow(self, material_name: str) -> Optional[float]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetBloodFlow', B, material_name, D)
        return None if not res[0] else res[1]

    def get_metabolic_rate(self, material_name: str) -> Optional[float]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetMetabolicRate', B, material_name, D)
        return None if not res[0] else res[1]

    def get_voxel_convection(self, material_name: str) -> Optional[float]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetVoxelConvection', B, material_name, D)
        return None if not res[0] else res[1]

    def exists(self, material_name: str) -> bool:
        return self.query_method_bool('Exists', material_name)

    def change_background_material(self):
        #NOTE: not documented, found in the history list
        self.record_method('ChangeBackgroundMaterial')

    def set_use_emissivity(self, flag: bool = True):
        #NOTE: not documented, found in the history list
        self.record_method('UseEmissivity', flag)


from . import Project
from . import ComObjectWrapper
from . import w32com
import numpy as np

class FarfieldPlot(ComObjectWrapper):
    PLOT_TYPE_POLAR = 'polar'
    PLOT_TYPE_CARTESIAN = 'cartesian'
    PLOT_TYPE_2D = '2d'
    PLOT_TYPE_2D_ORTHOGRAPHIC = '2dortho'
    PLOT_TYPE_3D = '3d'

    ASCII_VERSION_2009 = '2009'
    ASCII_VERSION_2010 = '2010'

    COORD_SYSTEM_SPHERICAL = 'spherical'
    COORD_SYSTEM_LUDWIG2AE = 'ludwig2ae'
    COORD_SYSTEM_LUDWIG2EA = 'ludwig2ea'
    COORD_SYSTEM_LUDWIG3 = 'ludwig3'

    POLARIZATION_LINEAR = 'linear'
    POLARIZATION_CIRCULAR = 'circular'
    POLARIZATION_SLANT = 'slant'
    POLARIZATION_ABS = 'abs'

    COMPONENT_RADIAL = 'radial'
    COMPONENT_COMP1 = 'comp1'
    COMPONENT_THETA = 'theta'
    COMPONENT_AZIMUTH = 'azimuth'
    COMPONENT_LEFT = 'left'
    COMPONENT_ALPHA = 'alpha'
    COMPONENT_HORIZONTAL = 'horizontal'
    COMPONENT_CROSSPOLAR = 'crosspolar'
    COMPONENT_COMP2 = 'comp2'
    COMPONENT_PHI = 'phi'
    COMPONENT_ELEVATION = 'elevation'
    COMPONENT_RIGHT = 'right'
    COMPONENT_EPSILON = 'epsilon'
    COMPONENT_VERTICAL = 'vertical'
    COMPONENT_COPOLAR = 'copolar'

    COMPLEX_COMP_ABS = 'abs'
    COMPLEX_COMP_PHASE = 'phase'
    COMPLEX_COMP_RE = 're'
    COMPLEX_COMP_IM = 'im'

    TF_TYPE_TIME = 'time'
    TF_TYPE_FREQUENCY = 'frequency'
    TF_TYPE_DEFAULT = ''

    LIST_COMPONENT_THETA = 'Point_T'
    LIST_COMPONENT_PHI = 'Point_P'
    LIST_COMPONENT_RADIUS = 'Point_R'

    CUT_TYPE_POLAR = 'polar'
    CUT_TYPE_LATERAL = 'lateral'

    PLOT_MODE_DIRECTIVITY = 'directivity'
    PLOT_MODE_GAIN = 'gain'
    PLOT_MODE_REALIZED_GAIN = 'realized gain'
    PLOT_MODE_EFIELD = 'efield'
    PLOT_MODE_EPATTERN = 'epattern'
    PLOT_MODE_HFIELD = 'hfield'
    PLOT_MODE_PFIELD = 'pfield'
    PLOT_MODE_RCS = 'rcs'
    PLOT_MODE_RCS_UNITS = 'rcsunits'
    PLOT_MODE_RCS_SW = 'rcssw'

    SEL_COMPONENT_ABS = 'Abs'
    SEL_COMPONENT_AXIAL_RATIO = 'Axial Ratio'
    SEL_COMPONENT_THETA = 'Theta'
    SEL_COMPONENT_THETA_PHASE = 'Theta Phase'
    SEL_COMPONENT_PHI = 'Phi'
    SEL_COMPONENT_PHI_PHASE = 'Phi Phase'
    SEL_COMPONENT_THETA_DIV_PHI = 'Theta/Phi'
    SEL_COMPONENT_PHI_DIV_THETA = 'Phi/Theta'

    # unit codes documented in CST Studio Suite Online Help
    #  (Automation and Scription -> Visual Basic (VBA) -> 3D Simulation VBA -> VBA Objects ->
    #  -> Post Processing -> FarfieldPlot -> function DBUnit
    UNIT_CODE_MINUS_1 = '-1'
    UNIT_CODE_0 = '0'
    UNIT_CODE_60 = '60'
    UNIT_CODE_120 = '120'
    UNIT_CODE_MINUS_60 = '-60'

    MAX_REF_MODE_ABS = 'abs'
    MAX_REF_MODE_PLOT = 'plot'

    AXES_TYPE_XYZ = 'xyz'
    AXES_TYPE_USER = 'user'
    AXES_TYPE_MAIN_LOBE = 'mainlobe'
    AXES_TYPE_CURRENT_WCS = 'currentwcs'

    ANTENNA_TYPE_UNKNOWN = 'unknown'
    ANTENNA_TYPE_ISOTROPIC = 'isotropic'
    ANTENNA_TYPE_ISOTROPIC_LINEAR = 'isotropic_linear'
    ANTENNA_TYPE_DIRECTIONAL_LINEAR = 'directional_linear'
    ANTENNA_TYPE_DIRECTIONAL_CIRCULAR = 'directional_circular'

    PHASE_CENTER_COMPONENT_THETA = 'theta'
    PHASE_CENTER_COMPONENT_PHI = 'phi'
    PHASE_CENTER_COMPONENT_BORESIGHT = 'boresight'

    PHASE_CENTER_PLANE_BOTH = 'both'
    PHASE_CENTER_PLANE_E = 'e-plane'
    PHASE_CENTER_PLANE_H = 'h-plane'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.FarfieldPlot

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def reset_plot(self):
        self.invoke_method('ResetPlot')

    def set_plot_type(self, plot_type: str):
        self.invoke_method('Plottype', plot_type)

    def vary_coord(self, angle_variant: int):
        # angle depends on active coordinate system:
        #   Coord system -> variant 1 | variant 2
        #   Spherical -> Theta | Phi
        #   Ludwig 2 AE -> Elevation | Azimuth
        #   Ludwig 2 EA -> Alpha | Epsilon
        #   Ludwig 3 -> Theta | Phi
        self.invoke_method('Vary', 'angle' + str(angle_variant))

    def set_phi_deg(self, angle_deg: float):
        self.invoke_method('Phi', angle_deg)

    def set_phi_rad(self, angle_rad: float):
        self.set_phi_deg(np.rad2deg(angle_rad))

    def set_theta_deg(self, angle_deg: float):
        self.invoke_method('Theta', angle_deg)

    def set_theta_rad(self, angle_rad: float):
        self.set_theta_deg(np.rad2deg(angle_rad))

    def set_theta_step_deg(self, angle_deg: float):
        self.invoke_method('Step', angle_deg)

    def set_theta_step_rad(self, angle_rad: float):
        self.set_theta_step_deg(np.rad2deg(angle_rad))

    def set_phi_step_deg(self, angle_deg: float):
        self.invoke_method('Step2', angle_deg)

    def set_phi_step_rad(self, angle_rad: float):
        self.set_phi_step_deg(np.rad2deg(angle_rad))

    def set_lock_steps(self, flag: bool = True):
        self.invoke_method('SetLockSteps', flag)

    def set_plot_range_only(self, flag: bool = True):
        self.invoke_method('SetPlotRangeOnly', flag)

    def set_theta_start_deg(self, angle_deg: float):
        self.invoke_method('SetThetaStart', angle_deg)

    def set_theta_start_rad(self, angle_rad: float):
        self.set_theta_start_deg(np.rad2deg(angle_rad))

    def set_theta_end_deg(self, angle_deg: float):
        self.invoke_method('SetThetaEnd', angle_deg)

    def set_theta_end_rad(self, angle_rad: float):
        self.set_theta_end_deg(np.rad2deg(angle_rad))

    def set_phi_start_deg(self, angle_deg: float):
        self.invoke_method('SetPhiStart', angle_deg)

    def set_phi_start_rad(self, angle_rad: float):
        self.set_phi_start_deg(np.rad2deg(angle_rad))

    def set_phi_end_deg(self, angle_deg: float):
        self.invoke_method('SetPhiEnd', angle_deg)

    def set_phi_end_rad(self, angle_rad: float):
        self.set_phi_end_deg(np.rad2deg(angle_rad))

    def use_farfield_approximation(self, flag: bool = True):
        self.invoke_method('UseFarfieldApproximation', flag)

    def set_multipoles_max_number(self, max_num: int):
        self.invoke_method('SetMultipolNumber', max_num)

    def set_frequency(self, freq: float):
        self.invoke_method('SetFrequency', freq)

    def set_time(self, time: float):
        self.invoke_method('SetTime', time)

    def set_time_domain_farfield(self, flag: bool = True):
        self.invoke_method('SetTimeDomainFF', flag)

    def set_num_movie_samples(self, num: int):
        self.invoke_method('SetMovieSamples', num)

    def plot(self):
        self.invoke_method('Plot')

    def store_settings(self):
        self.invoke_method('StoreSettings')

    def export_summary_as_ascii(self, file_name: str):
        self.invoke_method('ASCIIExportSummary', file_name)

    def set_ascii_export_version(self, version: str):
        self.invoke_method('ASCIIExportVersion', version)

    def export_source_as_ascii(self, file_name: str):
        self.invoke_method('ASCIIExportAsSource', file_name)

    def export_broadband_source_as_ascii(self, file_name: str):
        self.invoke_method('ASCIIExportAsBroadbandSource', file_name)

    def copy_farfield_to_1d_results(self, result_folder: str, result_name: str):
        self.invoke_method('CopyFarfieldTo1DResults', result_folder, result_name)

    def include_unit_cell_sidewalls(self, flag: bool = True):
        self.invoke_method('IncludeUnitCellSidewalls', flag)

    def calculate_point_deg(
            self, theta_deg: float, phi_deg: float, field_component: str, farfield_name: str = ''):
        return self.invoke_method(
            'CalculatePoint', theta_deg, phi_deg, field_component, farfield_name)

    def calculate_point_rad(
            self, theta_rad: float, phi_rad: float, field_component: str, farfield_name: str = ''):
        return self.calculate_point_deg(
            np.rad2deg(theta_rad), np.rad2deg(phi_rad), field_component, farfield_name)

    def calculate_point_no_approx_deg(
            self, theta_deg: float, phi_deg: float, radius: float, field_component: str,
            farfield_name: str = ''):
        return self.invoke_method(
            'CalculatePointNoApprox', theta_deg, phi_deg, radius, field_component, farfield_name)

    def calculate_point_no_approx_rad(
            self, theta_rad: float, phi_rad: float, radius: float, field_component: str,
            farfield_name: str = ''):
        return self.calculate_point_no_approx_deg(
            np.rad2deg(theta_rad), np.rad2deg(phi_rad), radius, field_component, farfield_name)

    def add_list_eval_point_deg(
            self, polar_angle_deg: float, lateral_angle_deg: float, radius: float,
            coord_system: str, tf_type: str, freq_or_time: float):
        self.invoke_method(
            'AddListEvaluationPoint', polar_angle_deg, lateral_angle_deg, radius,
            coord_system, tf_type, freq_or_time)

    def add_list_eval_point_rad(
            self, polar_angle_rad: float, lateral_angle_rad: float, radius: float,
            coord_system: str, tf_type: str, freq_or_time: float):
        self.add_list_eval_point_deg(
            np.rad2deg(polar_angle_rad), np.rad2deg(lateral_angle_rad),
            radius, coord_system, tf_type, freq_or_time)

    def calculate_list(self, name: str = ''):
        self.invoke_method('CalculateList', name)

    def get_list(self, field_component: str):
        return self.invoke_method('GetList', field_component)

    def get_list_item(self, index: int, field_component: str) -> float:
        return self.invoke_method('GetListItem', index, field_component)

    def clear_cuts(self):
        self.invoke_method('ClearCuts')

    def add_cut_deg(self, cut_type: str, const_angle_deg: float, step_deg: float):
        self.invoke_method('AddCut', cut_type, const_angle_deg, step_deg)

    def add_cut_rad(self, cut_type: str, const_angle_rad: float, step_rad: float):
        self.add_cut_deg(cut_type, np.rad2deg(const_angle_rad), np.rad2deg(step_rad))

    def set_color_by_value(self, flag: bool = True):
        self.invoke_method('SetColorByValue', flag)

    def set_theta_360_deg(self, flag: bool = True):
        self.invoke_method('SetTheta360', flag)

    def draw_step_lines(self, flag: bool = True):
        self.invoke_method('DrawStepLines', flag)

    def enable_symmetric_range(self, flag: bool = True):
        self.invoke_method('SymmetricRange', flag)

    def draw_iso_long_lat_lines(self, flag: bool = True):
        self.invoke_method('DrawIsoLongitudeLatitudeLines', flag)

    def show_structure(self, flag: bool = True):
        self.invoke_method('ShowStructure', flag)

    def show_structure_profile(self, flag: bool = True):
        self.invoke_method('ShowStructureProfile', flag)

    def set_structure_transparent(self, flag: bool = True):
        self.invoke_method('SetStructureTransparent', flag)

    def set_farfield_transparent(self, flag: bool = True):
        self.invoke_method('SetFarfieldTransparent', flag)

    def set_farfield_size(self, size: int):
        self.invoke_method('FarfieldSize', size)

    def set_plot_mode(self, plot_mode: str):
        self.invoke_method('SetPlotMode', plot_mode)

    def get_plot_mode(self):
        return self.invoke_method('GetPlotMode')

    def select_component(self, sel_component: str):
        self.invoke_method('SelectComponent', sel_component)

    def enable_polar_extra_lines(self):
        self.invoke_method('SetSpecials', 'enablepolarextralines')

    def disable_polar_extra_lines(self):
        self.invoke_method('SetSpecials', 'disablepolarextralines')

    def show_total_radiated_power_linear(self):
        self.invoke_method('SetSpecials', 'showtrp')

    def show_total_radiated_power_logarithmic(self):
        self.invoke_method('SetSpecials', 'showtrpdb')

    def hide_total_radiated_power(self):
        self.invoke_method('SetSpecials', 'showtrpoff')

    def show_total_isotropic_sensitivity_linear(self):
        self.invoke_method('SetSpecials', 'showtis')

    def show_total_isotropic_sensitivity_logarithmic(self):
        self.invoke_method('SetSpecials', 'showtisdb')

    def hide_total_isotropic_sensitivity(self):
        self.invoke_method('SetSpecials', 'showtisoff')

    def set_virtual_sphere_radius(self, radius: float):
        self.invoke_method('Distance', radius)

    def set_scale_linear(self):
        self.invoke_method('SetScaleLinear', True)

    def set_scale_logarithmic(self):
        self.invoke_method('SetScaleLinear', False)

    def is_scale_linear(self) -> bool:
        return self.invoke_method('IsScaleLinear')

    def is_scale_logarithmic(self) -> bool:
        return not self.is_scale_linear()

    def set_inverse_axial_ratio(self, flag: bool = True):
        self.invoke_method('SetInverseAxialRatio', flag)

    def has_inverse_axial_ratio(self) -> bool:
        return self.invoke_method('IsInverseAxialRatio')

    def set_log_plot_range(self, range_db: float):
        self.invoke_method('SetLogRange', range_db)

    def set_log_plot_normalization(self, norm_db: float):
        self.invoke_method('SetLogNorm', norm_db)

    def get_log_plot_range(self) -> float:
        return self.invoke_method('GetLogRange')

    def set_main_lobe_threshold(self, threshold_db: float):
        self.invoke_method('SetMainLobeThreshold', threshold_db)

    def set_db_unit(self, unit_code: str):
        self.invoke_method('DBUnit', unit_code)

    def set_max_reference_mode(self, max_ref_mode: str):
        self.invoke_method('SetMaxReferenceMode', max_ref_mode)

    def enable_fixed_plot_maximum(self, flag: bool = True):
        self.invoke_method('EnableFixPlotMaximum', flag)

    def is_plot_maximum_fixed(self) -> bool:
        return self.invoke_method('IsPlotMaximumFixed')

    def set_fixed_plot_maximum(self, maximum: float):
        self.invoke_method('SetFixPlotMaximumValue', maximum)

    def get_fixed_plot_maximum(self) -> float:
        return self.invoke_method('GetFixPlotMaximumValue')

    def set_num_contour_values(self, num: int):
        self.invoke_method('SetNumberOfContourValues', num)

    def draw_countour_lines(self, flag: bool = True):
        self.invoke_method('DrawContourLines', flag)

    def set_origin_by_bounding_box(self):
        self.invoke_method('Origin', 'bbox')

    def set_origin_to_zero(self):
        self.invoke_method('Origin', 'zero')

    def set_origin(self, x: float, y: float, z: float):
        self.invoke_method('Origin', 'free')
        self.invoke_method('Userorigin', x, y, z)

    def set_phi_start_axis(self, x: float, y: float, z: float):
        self.invoke_method('Phistart', x, y, z)

    def set_theta_start_axis(self, x: float, y: float, z: float):
        self.invoke_method('Thetastart', x, y, z)

    def set_axes_type(self, axes_type: str):
        self.invoke_method('SetAxesType', axes_type)

    def set_antenna_type(self, antenna_type: str):
        self.invoke_method('SetAntennaType', antenna_type)

    def set_polarization_vector(self, x: float, y: float, z: float):
        self.invoke_method('PolarizationVector', x, y, z)

    def set_coord_system(self, coord_system: str):
        self.invoke_method('SetCoordinateSystemType', coord_system)

    def set_automatic_coord_system(self, flag: bool = True):
        self.invoke_method('SetAutomaticCoordinateSystem', flag)

    def set_polarization_linear(self):
        self.invoke_method('SetPolarizationType', 'linear')

    def set_polarization_circular(self):
        self.invoke_method('SetPolarizationType', 'circular')

    def set_polarization_slant_deg(self, angle_deg: float):
        self.invoke_method('SetPolarizationType', 'slant')
        self.invoke_method('SlantAngle', angle_deg)

    def set_polarization_slant_rad(self, angle_rad: float):
        self.set_polarization_slant_deg(np.rad2deg(angle_rad))

    def use_decoupling_plane(self, flag: bool = True):
        self.invoke_method('UseDecouplingPlane', flag)

    def set_decoupling_plane_axis(self, axis: str):
        # axis: 'x' or 'y' or 'z'
        self.invoke_method('DecouplingPlaneAxis', axis)

    def set_decoupling_plane_position(self, position: float):
        self.invoke_method('DecouplingPlanePosition', position)

    def use_user_defined_decoupling_plane(self, flag: bool = True):
        self.invoke_method('SetUserDecouplingPlane', flag)

    def get_max_value(self) -> float:
        return self.invoke_method('Getmax')

    def get_min_value(self) -> float:
        return self.invoke_method('Getmin')

    def get_mean_value(self) -> float:
        return self.invoke_method('GetMean')

    def get_radiation_efficiency(self) -> float:
        return self.invoke_method('GetRadiationEfficiency')

    def get_total_efficiency(self) -> float:
        return self.invoke_method('GetTotalEfficiency')

    def get_system_radiation_efficiency(self) -> float:
        return self.invoke_method('GetSystemRadiationEfficiency')

    def get_system_total_efficiency(self) -> float:
        return self.invoke_method('GetSystemTotalEfficiency')

    def get_total_radiated_power(self) -> float: # in watts
        return self.invoke_method('GetTRP')

    def get_total_rcs(self) -> float:
        return self.invoke_method('GetTotalRCS')

    def get_total_acs(self) -> float:
        return self.invoke_method('GetTotalACS')

    def get_main_lobe_angle_deg(self) -> float:
        return self.invoke_method('GetMainLobeDirection')

    def get_main_lobe_angle_rad(self) -> float:
        return np.deg2rad(self.get_main_lobe_angle_deg())

    def get_main_lobe_vector(self):
        x = w32com.create_ref_double()
        y = w32com.create_ref_double()
        z = w32com.create_ref_double()
        self.invoke_method('GetMainLobeVector', x, y, z)
        return (x.value, y.value, z.value)

    def get_main_lobe_width_deg(self) -> float:
        return self.invoke_method('GetAngularWidthXdB')

    def get_main_lobe_width_rad(self) -> float:
        return np.deg2rad(self.get_main_lobe_width_deg())

    def get_side_lobe_suppression(self) -> float:
        return self.invoke_method('GetSideLobeSuppression')

    def get_side_lobe_level(self) -> float:
        return self.invoke_method('GetSideLobeLevel')

    def get_front_to_back_ratio(self) -> float:
        return self.invoke_method('GetFrontToBackRatio')

    def enable_phase_center_calculation(self, flag: bool = True):
        self.invoke_method('EnablePhaseCenterCalculation', flag)

    def set_phase_center_component(self, phase_center_component: str):
        self.invoke_method('SetPhaseCenterComponent', phase_center_component)

    def set_phase_center_plane(self, phase_center_plane: str):
        self.invoke_method('SetPhaseCenterPlane', phase_center_plane)

    def set_phase_center_angular_limit_deg(self, limit_deg: float):
        self.invoke_method('SetPhaseCenterAngularLimit', limit_deg)

    def set_phase_center_angular_limit_rad(self, limit_rad: float):
        self.set_phase_center_angular_limit_deg(np.rad2deg(limit_rad))

    def show_phase_center(self, flag: bool = True):
        self.invoke_method('ShowPhaseCenter', flag)

    def get_phase_center_avg(self):
        return self.__get_phase_center('avg')

    def get_phase_center_e_plane(self):
        return self.__get_phase_center('eplane')

    def get_phase_center_h_plane(self):
        return self.__get_phase_center('hplane')

    def get_phase_center_expr(self) -> str:
        return self.invoke_method('GetPhaseCenterResultExpr')

    def get_phase_center_expr_avg(self) -> str:
        return self.invoke_method('GetPhaseCenterResultExprAvg')

    def get_phase_center_expr_e_plane(self) -> str:
        return self.invoke_method('GetPhaseCenterResultExprEPlane')

    def get_phase_center_expr_h_plane(self) -> str:
        return self.invoke_method('GetPhaseCenterResultExprHPlane')

    def __get_phase_center(self, mode: str):
        x = self.invoke_method('GetPhaseCenterResult', 'x', mode)
        y = self.invoke_method('GetPhaseCenterResult', 'y', mode)
        z = self.invoke_method('GetPhaseCenterResult', 'z', mode)
        return (x, y, z)


from . import Shape, IVBAProvider

class Torus(Shape):
    AXIS_X = 'x'
    AXIS_Y = 'y'
    AXIS_Z = 'z'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Torus')

    def set_axis(self, axis: str):
        self.record_method('Axis', axis)

    def set_outer_radius(self, radius: float):
        self.record_method('Outerradius', radius)

    def set_inner_radius(self, radius: float):
        self.record_method('Innerradius', radius)

    def set_x_center(self, center: float):
        self.record_method('Xcenter', center)

    def set_y_center(self, center: float):
        self.record_method('Ycenter', center)

    def set_z_center(self, center: float):
        self.record_method('Zcenter', center)

    def set_num_segments(self, num_segments: int):
        self.record_method('Segments', num_segments)

    def set_smooth_geometry(self):
        self.set_num_segments(0)


__ALL__ = [

]

from .frequency_range_list import FrequencyRangeList
from .com_object_wrapper import ComObjectWrapper
from .interface import Interface
from .project import Project
from .project_units import ProjectUnits
from .material import Material
from .component import Component
from .group import Group
from .shape import Shape
from .analytical_face import AnalyticalFace
from .brick import Brick
from .cone import Cone
from .cylinder import Cylinder
from .elliptical_cylinder import EllipticalCylinder
from .sphere import Sphere
from .torus import Torus
from .wire import Wire
from .extrude import Extrude
from .rotate import Rotate
from .loft import Loft
from .trace_from_curve import TraceFromCurve
from .solid import Solid
from .align import Align
from .pick import Pick
from .transform import Transform
from .wcs import WCS
from .anchorpoint import Anchorpoint
from .curve import Curve
from .analytical_curve import AnalyticalCurve
from .arc import Arc
from .mesh_settings import MeshSettings
from .mesh import Mesh
from .mesh_adaptation_3d import MeshAdaption3D
from .mesh_shapes import MeshShapes
from .time_signal import TimeSignal
from .port import Port
from .discrete_port import DiscretePort
from .discrete_face_port import DiscreteFacePort
from .floquet_port import FloquetPort
from .plane_wave import PlaneWave
from .farfield_source import FarfieldSource
from .field_source import FieldSource
from .lumped_element import LumpedElement
from .lumped_face_element import LumpedFaceElement
from .monitor import Monitor
from .background import Background
from .layer_stacking import LayerStacking
from .boundary import Boundary
from .hf_solver import HFSolver
from .fd_solver import FDSolver
from .eigenmode_solver import EigenmodeSolver
from .ie_solver import IESolver
from .asymptotic_solver import AsymptoticSolver
from .optimizer import Optimizer
from .parameter_sweep import ParameterSweep
from .solver_parameter import SolverParameter
from .ads_cosimulation import ADSCosimulation
from .result_tree import ResultTree
from .result_tree_item import ResultTreeItem
from .farfield_plot import FarfieldPlot
from .result_1d import Result1D
from .result_1d_complex import Result1DComplex
from .post_process_1d import PostProcess1D
from .stl import STL
from .design import Design
from . import units


from . import Project
from . import ComObjectWrapper
from . import w32com

class DiscreteFacePort(ComObjectWrapper):
    TYPE_S_PARAMETER = 'Sparameter'
    TYPE_VOLTAGE = 'Voltage'
    TYPE_CURRENT = 'Current'

    FACE_TYPE_LINEAR = 'Linear'
    FACE_TYPE_CURVED = 'Curved'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.DiscreteFacePort

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def create(self):
        self.invoke_method('Create')

    def modify(self):
        self.invoke_method('Modify')

    def set_number(self, number: int):
        self.invoke_method('PortNumber', number)

    def set_label(self, label: str):
        self.invoke_method('Label', label)

    def set_type(self, port_type: str):
        self.invoke_method('Type', port_type)

    def set_impedance(self, impedance: float):
        self.invoke_method('Impedance', impedance)

    def set_voltage_amplitude(self, amplitude: float):
        self.invoke_method('VoltageAmplitude', amplitude)

    def set_voltage_port_impedance(self, impedance: float):
        self.invoke_method('VoltagePortImpedance', impedance)

    def set_current_amplitude(self, amplitude: float):
        self.invoke_method('CurrentAmplitude', amplitude)

    def set_current_port_admittance(self, admittance: float):
        self.invoke_method('CurrentPortAdmittance', admittance)

    def set_point1(self, x: float, y: float, z: float):
        self.invoke_method('SetP1', False, x, y, z)

    def set_pick_as_point1(self):
        self.invoke_method('SetP1', True, 0, 0, 0)

    def set_point2(self, x: float, y: float, z: float):
        self.invoke_method('SetP2', False, x, y, z)

    def set_pick_as_point2(self):
        self.invoke_method('SetP2', True, 0, 0, 0)

    def set_invert_direction(self, flag: bool = True):
        self.invoke_method('InvertDirection', flag)

    def set_local_coordinates(self, flag: bool = True):
        self.invoke_method('LocalCoordinates', flag)

    def set_center_edge(self, flag: bool = True):
        self.invoke_method('CenterEdge', flag)

    def set_use_projection(self, flag: bool = True):
        self.invoke_method('UseProjection', flag)

    def set_reverse_projection(self, flag: bool = True):
        self.invoke_method('ReverseProjection', flag)

    def set_lumped_element_name(self, name: str):
        self.invoke_method('LumpedElementName', name)

    def set_delete_lumped_element(self, flag: bool = True):
        self.invoke_method('DeleteLumpedElement', flag)

    def set_monitor(self, flag: bool = True):
        self.invoke_method('Monitor', flag)

    def set_allow_full_size(self, flag: bool = True):
        self.invoke_method('AllowFullSize', flag)

    def set_face_type(self, face_type: str):
        self.invoke_method('FaceType', face_type)

    def get_properties(self, port_number: int):
        port_type = w32com.create_ref_str()
        impedance = w32com.create_ref_double()
        current = w32com.create_ref_double()
        voltage = w32com.create_ref_double()
        voltage_port_impedance = w32com.create_ref_double()
        monitor = w32com.create_ref_bool()
        success = self.invoke_method(
            'GetProperties', port_number, port_type, impedance, current, voltage,
            voltage_port_impedance, monitor)
        return (
            port_type.value, impedance.value, current.value, voltage.value,
            voltage_port_impedance.value, monitor.value) if success else None

    def delete(self, port_number: int):
        self.project.get_port().delete(port_number)



from . import Modeler, Schematic, InterfacePCBS, Parameters

class IProject:
    def get_modeler(self) -> Modeler:
        pass

    def get_schematic(self) -> Schematic:
        pass

    def get_interface_pcbs(self) -> InterfacePCBS:
        pass

    def get_parameters(self) -> Parameters:
        pass

    def get_messages(self) -> list:
        pass

    def get_last_message_text(self) -> str:
        pass


from . import Project
from . import ComObjectWrapper

class Mesh(ComObjectWrapper):
    TYPE_PBA = 'PBA'
    TYPE_STAIRCASE = 'Staircase'
    TYPE_TETRAHEDRAL = 'Tetrahedral'
    TYPE_SURFACE = 'Surface'
    TYPE_SURFACE_ML = 'SurfaceML'

    PBA_TYPE_PBA = 'PBA'
    PBA_TYPE_FAST_PBA = 'Fast PBA'

    PARALLEL_MESHER_TYPE_HEX = 'Hex'
    PARALLEL_MESHER_TYPE_TET = 'Tet'

    PARALLEL_MESHER_MODE_MAXIMUM = 'maximum'
    PARALLEL_MESHER_MODE_CUSTOM = 'user-defined'
    PARALLEL_MESHER_MODE_NONE = 'none'

    AUTOMESH_REFINE_DIELECTRICS_NONE = 'None'
    AUTOMESH_REFINE_DIELECTRICS_WAVE = 'Wave'
    AUTOMESH_REFINE_DIELECTRICS_STATIC = 'Static'

    SURFACE_MESH_METHOD_GENERAL = 'General'
    SURFACE_MESH_METHOD_FAST = 'Fast'

    SURFACE_TOLERANCE_TYPE_RELATIVE = 'Relative'
    SURFACE_TOLERANCE_TYPE_ABSOLUTE = 'Absolute'

    VOLUME_MESH_METHOD_DELAUNAY = 'Delaunay'
    VOLUME_MESH_METHOD_ADVANCING_FRONT = 'Advancing Front'

    PICK_TYPE_MIDPOINT = 'Midpoint'
    PICK_TYPE_FACE_CENTER = 'Facecenter'
    PICK_TYPE_CENTERPOINT = 'Centerpoint'
    PICK_TYPE_ENDPOINT = 'Endpoint'
    PICK_TYPE_CIRCLE_ENDPOINT = 'CircleEndpoint'
    PICK_TYPE_CIRCLEPOINT = 'Circlepoint'

    SPATIAL_VAR_NONE = 'none'
    SPATIAL_VAR_SPHERICAL = 'spherical'
    SPATIAL_VAR_CURVE = 'curve'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Mesh

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def set_mesh_type(self, mesh_type: str):
        self.invoke_method('MeshType', mesh_type)

    def set_creator(self, creator: str):
        # undocumented, found in the history list
        self.invoke_method('SetCreator', creator)

    def set_connectivity_check(self, flag: bool = True):
        # undocumented, found in the history list
        self.invoke_method('ConnectivityCheck', flag)

    def set_cad_processing_method(self, method: str, undocumented_param: int = -1):
        # undocumented, found in the history list
        self.invoke_method('SetCADProcessingMethod', method, undocumented_param)

    def set_gpu_for_matrix_calculation_disabled(self, flag: bool = True):
        # undocumented, found in the history list
        self.invoke_method('SetGPUForMatrixCalculationDisabled', flag)

    def set_tst_version(self, version: str):
        self.invoke_method('TSTVersion', version)

    def set_pba_version(self, version: str):
        self.invoke_method('PBAVersion', version)

    def set_pba_type(self, pba_type: str):
        self.invoke_method('PBAType', pba_type)

    def set_automatic_pba_type(self, flag: bool = True):
        self.invoke_method('AutomaticPBAType', flag)

    def set_number_of_lines_per_wavelength(self, number: int):
        self.invoke_method('LinesPerWavelength', number)

    def set_min_number_of_steps(self, number: int):
        self.invoke_method('MinimumStepNumber', number)

    def set_ratio_limit(self, value: float):
        self.invoke_method('RatioLimit', value)

    def set_use_ratio_limit(self, flag: bool = True):
        self.invoke_method('UseRatioLimit', flag)

    def set_smallest_mesh_step(self, value: float):
        self.invoke_method('SmallestMeshStep', value)

    def set_steps_per_wavelength_tet(self, value: float):
        self.invoke_method('StepsPerWavelengthTet', value)

    def set_steps_per_wavelength_srf(self, value: float):
        self.invoke_method('StepsPerWavelengthSrf', value)

    def set_steps_per_wavelength_srf_ml(self, value: float):
        self.invoke_method('StepsPerWavelengthSrfML', value)

    def set_minimum_step_number_tet(self, value: float):
        self.invoke_method('MinimumStepNumberTet', value)

    def set_minimum_step_number_srf(self, value: float):
        self.invoke_method('MinimumStepNumberSrf', value)

    def set_minimum_step_number_srf_ml(self, value: float):
        self.invoke_method('MinimumStepNumberSrfML', value)

    def set_automesh(self, flag: bool = True):
        self.invoke_method('Automesh', flag)

    def set_material_refinement_tet(self, flag: bool = True):
        self.invoke_method('MaterialRefinementTet', flag)

    def set_equilibrate_mesh(self, flag: bool = True):
        self.invoke_method('EquilibrateMesh', flag)

    def set_equilibrate_mesh_ratio(self, value: float):
        self.invoke_method('EquilibrateMeshRatio', value)

    def set_use_cell_aspect_ratio(self, flag: bool = True):
        self.invoke_method('UseCellAspectRatio', flag)

    def set_cell_aspect_ratio(self, value: float):
        self.invoke_method('CellAspectRatio', value)

    def set_use_pec_edge_model(self, flag: bool = True):
        self.invoke_method('UsePecEdgeModel', flag)

    def set_point_acc_enhancement(self, value: float):
        self.invoke_method('PointAccEnhancement', value)

    def set_fast_pba_accuracy(self, value: float):
        self.invoke_method('FastPBAAccuracy', value)

    def set_fast_pba_gap_detection(self, flag: bool = True):
        self.invoke_method('FastPBAGapDetection', flag)

    def set_fast_pba_gap_tolerance(self, value: float):
        self.invoke_method('FPBAGapTolerance', value)

    def set_area_fill_limit(self, value: float):
        self.invoke_method('AreaFillLimit', value)

    def set_convert_geometry_data_after_meshing(self, flag: bool = True):
        self.invoke_method('ConvertGeometryDataAfterMeshing', flag)

    def set_consider_space_for_lower_mesh_limit(self, flag: bool = True):
        self.invoke_method('ConsiderSpaceForLowerMeshLimit', flag)

    def set_ratio_limit_governs_local_refinement(self, flag: bool = True):
        self.invoke_method('RatioLimitGovernsLocalRefinement', flag)

    def update(self):
        self.invoke_method('Update')

    def force_update(self):
        self.invoke_method('ForceUpdate')

    def calculate_matrices(self):
        self.invoke_method('CalculateMatrices')

    def set_view_mesh_mode(self, flag: bool = True):
        self.invoke_method('ViewMeshMode', flag)

    def set_small_feature_size(self, value: float):
        self.invoke_method('SmallFeatureSize', value)

    def set_parallel_mesher_mode(self, mesher_type: str, mesher_mode: str):
        self.invoke_method('SetParallelMesherMode', mesher_type, mesher_mode)

    def set_max_number_of_parallel_mesher_threads(self, mesher_type: str, number: int):
        self.invoke_method('SetMaxParallelMesherThreads', mesher_type, number)

    def set_automesh_straight_lines(self, flag: bool = True):
        self.invoke_method('AutomeshStraightLines', flag)

    def set_automesh_elliptical_lines(self, flag: bool = True):
        self.invoke_method('AutomeshEllipticalLines', flag)

    def enable_automesh_at_ellipse_bounds(self, factor: float):
        self.invoke_method('AutomeshAtEllipseBounds', True, factor)

    def disable_automesh_at_ellipse_bounds(self):
        self.invoke_method('AutomeshAtEllipseBounds', False, 0)

    def set_automesh_at_wire_end_points(self, flag: bool = True):
        self.invoke_method('AutomeshAtWireEndPoints', flag)

    def set_automesh_at_probe_points(self, flag: bool = True):
        self.invoke_method('AutomeshAtProbePoints', flag)

    def set_automesh_limit_shape_faces(self, flag: bool = True):
        self.invoke_method('AutoMeshLimitShapeFaces', flag)

    def set_automesh_number_of_shape_faces(self, number: int):
        self.invoke_method('AutoMeshNumberOfShapeFaces', number)

    def set_merge_thin_pec_layer_fixpoints(self, flag: bool = True):
        self.invoke_method('MergeThinPECLayerFixpoints', flag)

    def set_automesh_fixpoints_for_background(self, flag: bool = True):
        self.invoke_method('AutomeshFixpointsForBackground', flag)

    def enable_automesh_refine_at_pec_lines(self, factor: int):
        self.invoke_method('AutomeshRefineAtPecLines', True, factor)

    def disable_automesh_refine_at_pec_lines(self):
        self.invoke_method('AutomeshRefineAtPecLines', False, 0)

    def set_automesh_refine_pec_along_axes_only(self, flag: bool = True):
        self.invoke_method('AutomeshRefinePecAlongAxesOnly', flag)

    def set_automesh_refine_dielectrics_type(self, dielectrics_type: str):
        self.invoke_method('SetAutomeshRefineDielectricsType', dielectrics_type)

    def set_automesh_fixpoints_for_background(self, flag: bool = True):
        self.invoke_method('AutomeshFixpointsForBackground', flag)

    def set_surface_mesh_geometry_accuracy(self, value: float):
        self.invoke_method('SurfaceMeshGeometryAccuracy', value)

    def set_surface_mesh_method(self, method: str):
        self.invoke_method('SurfaceMeshMethod', method)

    def set_surface_tolerance(self, value: float):
        self.invoke_method('SurfaceTolerance', value)

    def set_surface_tolerance_type(self, surface_tolerance_type: str):
        self.invoke_method('SurfaceToleranceType', surface_tolerance_type)

    def set_normal_tolerance(self, value: float):
        self.invoke_method('NormalTolerance', value)

    def set_anisotropic_curvature_refinement_fsm(self, flag: bool = True):
        self.invoke_method('AnisotropicCurvatureRefinementFSM', flag)

    def set_surface_mesh_enrichment(self, level: int):
        self.invoke_method('SurfaceMeshEnrichment', level)

    def set_surface_optimization(self, flag: bool = True):
        self.invoke_method('SurfaceOptimization', flag)

    def set_surface_smoothing(self, flag: bool = True):
        self.invoke_method('SurfaceSmoothing', flag)

    def set_curvature_refinement_factor(self, value: float):
        self.invoke_method('CurvatureRefinementFactor', value)

    def set_min_curvature_refinement(self, value: float):
        self.invoke_method('MinimumCurvatureRefinement', value)

    def set_anisotropic_curvature_refinement(self, flag: bool = True):
        self.invoke_method('AnisotropicCurvatureRefinement', flag)

    def set_volume_optimization(self, flag: bool = True):
        self.invoke_method('VolumeOptimization', flag)

    def set_volume_smoothing(self, flag: bool = True):
        self.invoke_method('VolumeSmoothing', flag)

    def set_density_transitions(self, value: float):
        self.invoke_method('DensityTransitions', value)

    def set_volume_mesh_method(self, method: str):
        self.invoke_method('VolumeMeshMethod', method)

    def set_delaunay_optimization_level(self, value: int):
        self.invoke_method('DelaunayOptimizationLevel', value)

    def set_delaunay_propagation_factor(self, value: float):
        self.invoke_method('DelaunayPropagationFactor', value)

    def snap_to_surface_mesh(self, file_path_in: str, file_path_out: str):
        self.invoke_method('SnapToSurfaceMesh', file_path_in, file_path_out)

    def set_self_intersecting_check(self, flag: bool = True):
        self.invoke_method('SelfIntersectingCheck', flag)

    def find_fixpoint_from_position(self, x: float, y: float, z: float) -> int:
        return self.invoke_method('FindFixpointFromPosition', x, y, z)

    def add_fixpoint(self, x: float, y: float, z: float):
        self.invoke_method('AddFixpoint', x, y, z)

    def add_fixpoint_relative(self, ref_fixpoint_id: int, x: float, y: float, z: float):
        self.invoke_method('RelativeAddFixpoint', ref_fixpoint_id, x, y, z)

    def delete_fixpoint(self, fixpoint_id: int):
        self.invoke_method('DeleteFixpoint', fixpoint_id)

    def add_intermediate_fixpoints(self, id1: int, id2: int, number_of_points: int):
        self.invoke_method('AddIntermediateFixpoint', id1, id2, number_of_points)

    def add_automesh_fixpoint(
            self, use_x: bool, use_y: bool, use_z: bool, x: float, y: float, z: float):
        self.invoke_method('AddAutomeshFixpoint', use_x, use_y, use_z, x, y, z)

    def delete_automesh_fixpoint(self, x: float, y: float, z: float):
        self.invoke_method('DeleteAutomeshFixpoint', x, y, z)

    def modify_automesh_fixpoint(self, use_x: bool, use_y: bool, use_z: bool, fixpoint_id: int):
        self.invoke_method('ModifyAutomeshFixpointFromId', use_x, use_y, use_z, fixpoint_id)

    def add_automesh_fixpoints_from_pick(
            self, use_x: bool, use_y: bool, use_z: bool, pick_type: str, solid_name: str,
            pick_id: int, face_id: int, number: int):
        self.invoke_method(
            'AddAutomeshFixpointFromId', use_x, use_y, use_z, pick_type, solid_name,
            pick_id, face_id, number)

    def delete_automesh_fixpoint(self, fixpoint_id: int):
        self.invoke_method('DeleteAutomeshFixpointFromId', fixpoint_id)

    def clear_spatial_variation(self):
        self.invoke_method('ClearSpatialVariation')

    def clear_spatial_variation_for_shape(self, solid_name: str):
        self.invoke_method('ClearSpatialVariationForShape', solid_name)

    def set_spatial_variation_type_for_shape(self, solid_name: str, var_type: str):
        self.invoke_method('SetSpatialVariationTypeForShape', solid_name, var_type)